#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## minizinc.py
##
"""
    Interface to MiniZinc's Python API.

    MiniZinc is a free and open-source constraint modeling language.
    MiniZinc is used to model constraint satisfaction and optimization problems in
    a high-level, solver-independent way, taking advantage of a large library of
    pre-defined constraints. The model is then compiled into FlatZinc, a solver input
    language that is understood by a wide range of solvers.
    https://www.minizinc.org

    The MiniZinc interface is text-based: CPMpy writes a textfile and passes it to the minizinc Python package.

    Always use :func:`cp.SolverLookup.get("minizinc") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'minizinc' python package is installed:

    .. code-block:: console

        $ pip install minizinc

    as well as the MiniZinc bundled binary packages, downloadable from:
    https://www.minizinc.org/software.html

    See detailed installation instructions at:
    https://minizinc-python.readthedocs.io/en/latest/getting_started.html

    Note for **Jupyter notebook** users: MiniZinc uses AsyncIO, so using it in a Jupyter notebook gives
    you the following error: ``RuntimeError: asyncio.run() cannot be called from a running event loop``
    You can overcome this by ``pip install nest_asyncio``
    and adding in the top cell ``import nest_asyncio; nest_asyncio.apply()``

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_minizinc

    ==============
    Module details
    ==============
"""
import re
from typing import Optional
import warnings
import sys
import os
import json
from datetime import timedelta  # for mzn's timeout

import numpy as np
import pkg_resources

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..exceptions import MinizincNameException, MinizincBoundsException
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.python_builtins import any as cpm_any
from ..expressions.variables import _NumVarImpl, _IntVarImpl, _BoolVarImpl, NegBoolView, cpm_array
from ..expressions.globalconstraints import DirectConstraint
from ..expressions.utils import is_num, is_any_list, argvals, argval
from ..transformations.decompose_global import decompose_in_tree
from ..exceptions import MinizincPathException, NotSupportedError
from ..transformations.get_variables import get_variables
from ..transformations.normalize import toplevel_list


class CPM_minizinc(SolverInterface):
    """
    Interface to MiniZinc's Python API

    Creates the following attributes (see parent constructor for more):

    - ``mzn_model``: object, the minizinc.Model instance
    - ``mzn_solver``: object, the minizinc.Solver instance
    - ``mzn_txt_solve``: str, the 'solve' item in text form, so it can be overwritten
    - ``mzn_result``: object, containing solve results

    The :class:`~cpmpy.expressions.globalconstraints.DirectConstraint`, when used, adds a constraint with that name and the given args to the MiniZinc model.

    Documentation of the solver's own Python API:
    https://minizinc-python.readthedocs.io/
    """

    required_version = (2, 8, 2)

    @staticmethod
    def supported(): 
        return CPM_minizinc.installed() and CPM_minizinc.executable_installed() and not CPM_minizinc.outdated()

    @staticmethod
    def installed():
        # try to import the package
        try:
            #  check if MiniZinc Python is installed
            import minizinc
            return True
        except ModuleNotFoundError:
            return False
        except Exception as e:
            raise e

    @staticmethod
    def executable_installed():
        # check if MiniZinc executable is installed
        from minizinc import default_driver
        if default_driver is None:
            warnings.warn("MiniZinc Python is installed, but the MiniZinc executable is missing in path.")
            return False
        return True

    @staticmethod
    def outdated():
        from minizinc import default_driver
        if default_driver.parsed_version >= CPM_minizinc.required_version:
            return False
        else:
            # outdated
            return True
                

    @staticmethod
    def solvernames(installed:bool=True, with_version:bool=False):
        """
            Returns solvers supported by MiniZinc (on your system) with optionally their installed version.

            Arguments:
                installed (boolean): whether to filter the solvernames to those installed on your system (default True)
                with_version (boolean): whether to additionally return the available version matching with each solvername 
                                        (if not available on the system, the entry defaults to None)

            Returns:
                list of solver names if with_version==False, otherwise a tuple of two lists: the solver names and their versions

            .. warning::
                WARNING, some of the returned solver names (when ``installed=False``) may not actually 
                be installed on your system (namely cplex, gurobi, scip, xpress).
                The following are bundled with minizinc: chuffed, coin-bc, gecode.
                Use ``installed=True`` (the default) if you only want the names of the actually installed solvers.
        """

        # Collect solver names and versions
        if CPM_minizinc.supported(): # check if minizinc is installed
            import minizinc
            driver = minizinc.default_driver
            
            # Collect solver names
            all_solvers, all_versions = [], []
            output = driver._run(["--solvers-json"]) # get json-structured solver overview
            solvers = json.loads(output.stdout)        
            for solver_dict in solvers:
                # get subsolver metadata
                tag = solver_dict["id"].split(".")[-1]
                version = solver_dict["version"]
                if tag not in ['findmus', 'gist', 'globalizer']: # some are not actually solvers
                    if tag not in all_solvers:
                        all_solvers.append(tag)
                        if version == '<unknown version>': # if no version info available, default to None
                            version = None
                        all_versions.append(version)
        else:
            warnings.warn("MiniZinc is not installed or not supported on this system.")
            if with_version:
                return ([], [])
            else:
                return []
            

        if not installed:
            """
            Return all solver names, without checking if they're actually available on the system.
            """
            if with_version:
                return (all_solvers, all_versions)
            else:
                return all_solvers

        else:
            """
            Test which solver executables are available by retrieving version information
            """
            valid_indices = [i for i, version in enumerate(all_versions) if version != "<unknown version>"] # check if version is available (required by minizinc)
            installed_solvers = [all_solvers[i] for i in valid_indices]
            installed_versions = [all_versions[i] for i in valid_indices]

            if with_version:
                return (installed_solvers, installed_versions)
            else:
                return installed_solvers       
             
    @staticmethod
    def solverversion(subsolver:str) -> Optional[str]:
        """
        Returns the version of the requested subsolver.

        Arguments:
            subsolver (str): name of the subsolver

        Returns:
            Version number of the subsolver if installed, else None 
        """
        all_solvers, all_versions = CPM_minizinc.solvernames(installed=False, with_version=True)
        try:
            solver_index = all_solvers.index(subsolver) # find requested subsolver
            return all_versions[solver_index]
        except ValueError:
            raise ValueError(f"Subsolver '{subsolver}' not found in the list of available solvers.")

    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.

        For Minizinc, two version numbers get returned: ``<minizinc python API version>/<minizinc driver version>``
        """
        try:
            from minizinc import default_driver
            return f"{pkg_resources.get_distribution('minizinc').version}/{'.'.join(str(a) for a in default_driver.parsed_version)}"
        except (pkg_resources.DistributionNotFound, ModuleNotFoundError):
            return None

    # variable name can not be any of these keywords
    keywords = frozenset(['ann', 'annotation', 'any', 'array', 'bool', 'case', 'constraint', 'diff', 'div', 'else',
                          'elseif', 'endif', 'enum', 'false', 'float', 'function', 'if', 'in', 'include', 'int',
                          'intersect', 'let', 'list', 'maximize', 'minimize', 'mod', 'not', 'of', 'op', 'opt', 'output',
                          'par', 'predicate', 'record', 'satisfy', 'set', 'solve', 'string', 'subset', 'superset',
                          'symdiff', 'test', 'then', 'true', 'tuple', 'type', 'union', 'var', 'where', 'xor'])
    # variable names must have this pattern
    mzn_name_pattern = re.compile('^[A-Za-z][A-Za-z0-9_]*$')

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
            cpm_model: Model(), a CPMpy Model() (optional)
            subsolver: str, name of a subsolver (optional)
                          has to be one of solvernames()
        """
        if not self.installed():
            raise Exception("CPM_minizinc: Install the python package 'minizinc' to use this solver interface.")
        elif not self.executable_installed():
            raise Exception("CPM_minizinc: Install the MiniZinc executable and make it available in path.")
        elif self.outdated():
            version = str(self.required_version[0])
            for x in self.required_version[1:]:
                version = version + "." + str(x)
            raise ImportError("Your Minizinc compiler is outdated, please upgrade to a version >= " + version)

        import minizinc

        # determine subsolver
        if subsolver is None or subsolver == 'minizinc':
            # default solver
            subsolver = "gecode"
        elif subsolver.startswith('minizinc:'):
            subsolver = subsolver[9:]  # strip 'minizinc:'

        # initialise the native solver object
        # (so its params can still be changed before calling solve)
        self.mzn_solver = minizinc.Solver.lookup(subsolver)
        # It is the model object that contains the constraints for minizinc
        self.mzn_model = minizinc.Model()
        self.mzn_model.add_string("% Generated by CPMpy\ninclude \"globals.mzn\";\n\n")
        # Prepare solve statement, so it can be overwritten on demand
        self.mzn_txt_solve = "solve satisfy;"
        self.mzn_result = None

        # initialise everything else and post the constraints/objective
        super().__init__(name="minizinc:"+subsolver, cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.mzn_model


    def _pre_solve(self, time_limit=None, **kwargs):
        """ shared by solve() and solveAll() """
        import minizinc

        if time_limit is not None:
            kwargs['timeout'] = timedelta(seconds=time_limit)

        # hack, we need to add the objective in a way that it can be changed
        # later, so make copy of the mzn_model
        copy_model = self.mzn_model.__copy__()  # it is implemented
        copy_model.add_string(self.mzn_txt_solve)
        # Transform Model into an instance
        mzn_inst = minizinc.Instance(self.mzn_solver, copy_model)

        kwargs['output-time'] = True  # required for time getting
        return (kwargs, mzn_inst)

    def solve(self, time_limit=None, **kwargs):
        """
            Call the MiniZinc solver
            
            Creates and calls an Instance with the already created ``mzn_model`` and ``mzn_solver``

            Arguments:
                time_limit (float, optional):       maximum solve time in seconds 
                **kwargs (any keyword argument):    sets parameters of solver object
                
            
            Arguments that correspond to solver parameters:

            =======================  ===========
            Keyword                  Description
            =======================  ===========
            free_search=True              Allow the solver to ignore the search definition within the instance. (Only available when the -f flag is supported by the solver). (Default: 0)
            optimisation_level=0          Set the MiniZinc compiler optimisation level. (Default: 1; 0=none, 1=single pass, 2=double pass, 3=root node prop, 4,5=probing)
            =======================  ===========             
            
            
            I am not sure where solver-specific arguments are documented, but the docs say that command line arguments can be passed by ommitting the '-' (e.g. 'f' instead of '-f')?
            
            The minizinc solver parameters are partly defined in its API:
            https://minizinc-python.readthedocs.io/en/latest/api.html#minizinc.instance.Instance.solve

            Does not store the ``minizinc.Instance()`` or ``minizinc.Result()``
        """

        if time_limit is not None and time_limit <= 0:
            raise ValueError("Time limit must be positive")

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # make mzn_inst
        (mzn_kwargs, mzn_inst) = self._pre_solve(time_limit=time_limit, **kwargs)
        
        # call the solver, with parameters
        import minizinc.error
        try:
            self.mzn_result = mzn_inst.solve(**mzn_kwargs)
        except minizinc.error.MiniZincError as e:
            if sys.platform == "win32" or sys.platform == "cygwin":  # path error can occur in windows
                path = os.environ.get("path")
                if "MiniZinc" in str(path):
                    warnings.warn('You might have the wrong minizinc PATH set (windows user Environment Variables')
                    raise e
                else:
                    raise MinizincPathException("Please add your minizinc installation folder to the user Environment PATH variable")
            else:
                raise e
        # new status, translate runtime
        self.cpm_status = self._post_solve(self.mzn_result)

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:  # mzn_result.status.has_solution():
            mznsol = self.mzn_result.solution
            if is_any_list(mznsol):
                print("Warning: multiple solutions found, only returning last one")
                mznsol = mznsol[-1]

            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if hasattr(mznsol, sol_var):
                    cpm_var._value = getattr(mznsol, sol_var)
                else:
                    raise ValueError(f"Var {cpm_var} is unknown to the Minizinc solver, this is unexpected - please report on github...")

            # translate objective, for optimisation problems only (otherwise None)
            self.objective_value_ = self.mzn_result.objective

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var._value = None

        return has_sol

    def _post_solve(self, mzn_result, solve_all:bool=False):
        """ shared by solve() and solveAll() """
        import minizinc

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        runtime = 0
        if 'time' in mzn_result.statistics:
            self.cpm_status.runtime = self.mzn_time_to_seconds(mzn_result.statistics.get("time"))
        else:
            runtime += self.mzn_time_to_seconds(mzn_result.statistics.get("flatTime", 0))
            runtime += self.mzn_time_to_seconds(mzn_result.statistics.get("initTime", 0))
            runtime += self.mzn_time_to_seconds(mzn_result.statistics.get("solveTime", 0))
            if runtime != 0:
                self.cpm_status.runtime = runtime
            else:
                raise NotImplementedError  # Please report on github, minizinc probably changed their time names/types

        # translate exit status
        mzn_status = mzn_result.status
        if mzn_status == minizinc.result.Status.SATISFIED:
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif mzn_status == minizinc.result.Status.ALL_SOLUTIONS:
            if solve_all:
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif mzn_status == minizinc.result.Status.OPTIMAL_SOLUTION:
            self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        elif mzn_status == minizinc.result.Status.UNSATISFIABLE:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif mzn_status == minizinc.result.Status.ERROR:
            self.cpm_status.exitstatus = ExitStatus.ERROR
            raise Exception("MiniZinc solver returned with status 'Error'")
        elif mzn_status == minizinc.result.Status.UNKNOWN:
            # means, no solution was found (e.g. within timeout?)...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else:
            raise NotImplementedError  # a new status type was introduced, please report on github

        return self.cpm_status

    def mzn_time_to_seconds(self, time):
        if isinstance(time, int):
            return time / 1000
        elif isinstance(time, timedelta):
            return time.total_seconds()  # --output-time
        else:
            raise NotImplementedError  # unexpected type for time

    async def _solveAll(self, display=None, time_limit=None, solution_limit=None, **kwargs):
        """ Special 'async' function because mzn.solutions() is async """

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))
        
        # make mzn_inst
        (kwargs, mzn_inst) = self._pre_solve(time_limit=time_limit, **kwargs)
        kwargs['all_solutions'] = True

        solution_count = 0
        # has an asynchronous generator
        async for mzn_result in mzn_inst.solutions(**kwargs):
            # was the last one
            if mzn_result.solution is None:
                break

             # fill in variable values
            mznsol = mzn_result.solution
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if hasattr(mznsol, sol_var):
                    cpm_var._value = getattr(mznsol, sol_var)
                else:
                    raise ValueError(f"Var {cpm_var} is unknown to the Minizinc solver, this is unexpected - please report on github...")

            # display if needed
            if display is not None:
                if isinstance(display, Expression):
                    print(argval(display))
                elif isinstance(display, list):
                    print(argvals(display))
                else:
                    display()  # callback

            # count and stop
            solution_count += 1
            if solution_count == solution_limit:
                break

            # add nogood on the user variables
            self += cpm_any([v != v.value() for v in self.user_vars])

        if solution_count == 0:
            # clear user vars if no solution found
            self.objective_value_ = None
            for var in self.user_vars:
                var._value = None

        # status handling
        self._post_solve(mzn_result, solve_all=True)

        if solution_count: # found at least one solution
            if solution_count == solution_limit: # matched solution limit
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            # elif mzn_result.solution is None: <- is implicit since nothing needs to update
                # last iteration didn't find a solution
                # nothing needs to update since _post_solve already set state correctly (state from the second-last iteration)

        return solution_count

    def solver_var(self, cpm_var) -> str:
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created.

            Returns:
                minizinc-friendly 'string' name of var.

            .. warning::
                WARNING, this assumes it is never given a 'NegBoolView'
                might not be true... e.g. in revar after solve?
        """
        if is_num(cpm_var):
            if cpm_var < -2147483646 or cpm_var > 2147483646:
                raise MinizincBoundsException(
                    "minizinc does not accept integer literals with bounds outside of range (-2147483646..2147483646)")
            return str(cpm_var)

        # Assumes it is never given a 'NegBoolView'
        if isinstance(cpm_var, NegBoolView):
            raise NotSupportedError("Negative literals are not handled here. Please report.")

        if cpm_var not in self._varmap:
            # clean the varname
            varname = cpm_var.name
            mzn_var = varname.replace(',', '_').replace('.', '_').replace(' ', '_').replace('[', '_').replace(']', '')

            # test if the name is a valid minizinc identifier
            if not self.mzn_name_pattern.search(mzn_var):
                raise MinizincNameException("Minizinc only accept names with alphabetic characters, "
                                            "digits and underscores. "
                                "First character must be an alphabetic character")
            if mzn_var in self.keywords:
                raise MinizincNameException(f"This variable name is a disallowed keyword in MiniZinc: {mzn_var}")

            if isinstance(cpm_var, _BoolVarImpl):
                self.mzn_model.add_string(f"var bool: {mzn_var};\n")
            elif isinstance(cpm_var, _IntVarImpl):
                if cpm_var.lb < -2147483646 or cpm_var.ub > 2147483646:
                    raise MinizincBoundsException("minizinc does not accept variables with bounds outside "
                                                  "of range (-2147483646..2147483646)")
                self.mzn_model.add_string(f"var {cpm_var.lb}..{cpm_var.ub}: {mzn_var};\n")
            self._varmap[cpm_var] = mzn_var

        return self._varmap[cpm_var]

    def objective(self, expr, minimize):
        """
            Post the given expression to the solver as objective to minimize/maximize

            - expr: Expression, the CPMpy expression that represents the objective function
            - minimize: Bool, whether it is a minimization problem (True) or maximization problem (False)

            'objective()' can be called multiple times, only the last one is stored
        """
        # get_variables(expr, collect=self.user_vars)  # add objvars to vars  # all are user vars

        # make objective function or variable and post
        obj = self._convert_expression(expr)
        # do not add it to the mzn_model yet, supports only one 'solve' entry
        if minimize:
            self.mzn_txt_solve = "solve minimize {};\n".format(obj)
        else:
            self.mzn_txt_solve = "solve maximize {};\n".format(obj)

    def has_objective(self):
        return self.mzn_txt_solve != "solve satisfy;"

    def transform(self, cpm_expr):
        """
            Decompose globals not supported (and flatten globalfunctions)
            ensure it is a list of constraints

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        cpm_cons = toplevel_list(cpm_expr)
        supported = {"min", "max", "abs", "element", "count", "nvalue", "alldifferent", "alldifferent_except0", "allequal",
                     "inverse", "ite" "xor", "table", "cumulative", "circuit", "gcc", "increasing", "decreasing",
                     "precedence", "no_overlap",
                     "strictly_increasing", "strictly_decreasing", "lex_lesseq", "lex_less", "lex_chain_less", 
                     "lex_chain_lesseq", "among"}
        return decompose_in_tree(cpm_cons, supported, supported_reified=supported - {"circuit", "precedence"}, csemap=self._csemap)

    def add(self, cpm_expr):
        """
            Translate a CPMpy constraint to MiniZinc string and add it to the solver

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
        """
        get_variables(cpm_expr, collect=self.user_vars)
        # transform and post the constraints
        for cpm_con in self.transform(cpm_expr):
            # Get text expression, add to the solver
            mzn_str = f"constraint {self._convert_expression(cpm_con)};\n"
            self.mzn_model.add_string(mzn_str)

        return self
    __add__ = add  # avoid redirect in superclass

    def _convert_expression(self, expr) -> str:
        """
            Convert a CPMpy expression into a minizinc-compatible string

            recursive: also converts nested subexpressions, so we need a
            function that returns strings
        """
        if is_any_list(expr):
            if isinstance(expr, np.ndarray):
                # must flatten
                expr_str = [self._convert_expression(e) for e in expr.flat]
            else:
                expr_str = [self._convert_expression(e) for e in expr]
            return "[{}]".format(",".join(expr_str))

        if isinstance(expr, (bool, np.bool_)):
            expr = BoolVal(expr)

        if not isinstance(expr, Expression):
            return self.solver_var(expr)  # constants

        if isinstance(expr, BoolVal):
            return str(expr.args[0]).lower()

        # default
        if isinstance(expr, _NumVarImpl):
            if isinstance(expr, NegBoolView):
                return "not " + self.solver_var(expr._bv)
            return self.solver_var(expr)

        # table(vars, tbl): no [] nesting of args, and special table output...
        if expr.name == "table":
            str_vars = self._convert_expression(expr.args[0])
            str_tbl = "[|\n"  # opening
            for row in expr.args[1]:
                str_tbl += ",".join(map(str, row)) + " |"  # rows
            str_tbl += "\n|]"  # closing
            return "table({}, {})".format(str_vars, str_tbl)

        # inverse(fwd, rev): unpack args and work around MiniZinc's default 1-based indexing
        if expr.name == "inverse":
            def zero_based(array):
                return "array1d(0..{}, {})".format(len(array)-1, self._convert_expression(array))

            str_fwd = zero_based(expr.args[0])
            str_rev = zero_based(expr.args[1])
            return "inverse({}, {})".format(str_fwd, str_rev)

        if expr.name == "alldifferent_except0":
            args_str = [self._convert_expression(e) for e in expr.args]
            return "alldifferent_except_0([{}])".format(",".join(args_str))

        if expr.name in ["lex_lesseq", "lex_less"]:
            X = [self._convert_expression(e) for e in expr.args[0]]
            Y = [self._convert_expression(e) for e in expr.args[1]]
            return f"{expr.name}({{}}, {{}})".format(X, Y)

        if expr.name in ["lex_chain_less", "lex_chain_lesseq"]:
            X = cpm_array([[self._convert_expression(e) for e in row] for row in expr.args])
            str_X = "[|\n"  # opening
            for row in X.T:  # Minizinc enforces lexicographic order on columns
                str_X += ",".join(map(str, row)) + " |"  # rows
            str_X += "\n|]"  # closing
            return f"{expr.name}({{}})".format(str_X)

        args_str = [self._convert_expression(e) for e in expr.args]
        # standard expressions: comparison, operator, element
        if isinstance(expr, Comparison):
            # wrap args that are a subexpression in ()
            for i, arg_str in enumerate(args_str):
                if isinstance(expr.args[i], Expression):  # (Comparison, Operator)
                    args_str[i] = "(" + args_str[i] + ")"
            # infix notation
            return "{} {} {}".format(args_str[0], expr.name, args_str[1])

        elif isinstance(expr, Operator):
            # some names differently (the infix names!)
            printmap = {'and': '/\\', 'or': '\\/',
                        'sum': '+', 'sub': '-',
                        'mul': '*', 'pow': '^'}
            op_str = expr.name
            expr_bounds = expr.get_bounds()
            if expr_bounds[0] < -2147483646 or expr_bounds[1] > 2147483646:
                raise MinizincBoundsException("minizinc does not accept expressions with bounds outside of "
                                              "range (-2147483646..2147483646)")
            if op_str in printmap:
                op_str = printmap[op_str]

            # TODO: pretty printing of () as in Operator?

            # special case: unary -
            if self.name == '-':
                return "-{}".format(args_str[0])

            # very special case: weighted sum (before 2-ary)
            if expr.name == 'wsum':
                # I don't think there is a more direct way unfortunately
                w = [self._convert_expression(wi) for wi in expr.args[0]]
                x = [self._convert_expression(xi) for xi in expr.args[1]]
                args_str = [f"{wi}*({xi})" for wi, xi in zip(w, x)]
                return "{}([{}])".format("sum", ",".join(args_str))

            # special case, infix: two args
            if len(args_str) == 2:
                # wrap args that are a subexpression in ()
                for i, arg_str in enumerate(args_str):
                    if isinstance(expr.args[i], Expression):
                        args_str[i] = "(" + args_str[i] + ")"
                return "{} {} {}".format(args_str[0], op_str, args_str[1])

            # special case: n-ary (non-binary): rename operator
            printnary = {'and': 'forall', 'or': 'exists', 'sum': 'sum'}
            if expr.name in printnary:
                op_str = printnary[expr.name]
                return "{}([{}])".format(op_str, ",".join(args_str))

            # default: prefix printing
            return "{}({})".format(op_str, ",".join(args_str))

        elif expr.name == "element":
            subtype = "int"
            if all(isinstance(v, bool) or \
                   (isinstance(v, Expression) and v.is_bool()) \
                   for v in expr.args[0]):
                subtype = "bool"
            idx = args_str[1]
            # minizinc is offset 1, which can be problematic for element...
            # thx Hakan, fix by using array1d(0..len, []), issue #54
            txt = "\n    let {{ array[int] of var {}: arr=array1d({}..{},{}) }} in\n".format(subtype, 0,
                                                                                             len(expr.args[0]) - 1,
                                                                                             args_str[0])
            txt += f"      arr[{idx}]"
            return txt

        # rest: global constraints
        elif expr.name.endswith('circuit'):  # circuit, subcircuit
            # minizinc is offset 1, which can be problematic here...
            args_str = ["{}+1".format(self._convert_expression(e)) for e in expr.args]

        elif expr.name == "cumulative":
            start, dur, end, _, _ = expr.args

            durstr = self._convert_expression([s + d == e for s, d, e in zip(start, dur, end)])
            format_str = "forall(" + durstr + " ++ [cumulative({},{},{},{})])"

            return format_str.format(args_str[0], args_str[1], args_str[3], args_str[4])

        elif expr.name == "precedence":
            return "value_precede_chain({},{})".format(args_str[1], args_str[0])

        elif expr.name == "no_overlap":
            start, dur, end = expr.args
            durstr = self._convert_expression([s + d == e for s, d, e in zip(start, dur, end)])
            format_str = "forall(" + durstr + " ++ [disjunctive({},{})])"
            return format_str.format(args_str[0], args_str[1])

        elif expr.name == 'ite':
            cond, tr, fal = expr.args
            return "if {} then {} else {} endif".format(self._convert_expression(cond), self._convert_expression(tr),
                                                        self._convert_expression(fal))

        elif expr.name == "gcc":
            vars, vals, occ = expr.args
            vars = self._convert_expression(vars)
            vals = self._convert_expression(vals)
            occ = self._convert_expression(occ)
            if expr.closed is False:
                name = "global_cardinality"
            else:
                name = "global_cardinality_closed"
            return "{}({},{},{})".format(name, vars, vals, occ)

        elif expr.name == "abs":
            return "abs({})".format(args_str[0])

        elif expr.name == "count":
            vars, val = expr.args
            vars = self._convert_expression(vars)
            val = self._convert_expression(val)
            return "count({},{})".format(vars, val)

        elif expr.name == "among":
            vars, vals = expr.args
            vars = self._convert_expression(vars)
            vals = self._convert_expression(vals).replace("[", "{").replace("]", "}")  # convert to set
            return "among({},{})".format(vars, vals)

        # a direct constraint, treat differently for MiniZinc, a text-based language
        # use the name as, unpack the arguments from the argument tuple
        elif isinstance(expr, DirectConstraint):
            return "{}({})".format(expr.name, ",".join(args_str))

        print_map = {"allequal": "all_equal", "xor": "xorall"}
        if expr.name in print_map:
            return "{}([{}])".format(print_map[expr.name], ",".join(args_str))

        # default (incl name-compatible global constraints...)
        return "{}([{}])".format(expr.name, ",".join(args_str))

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            Compute all solutions and optionally display the solutions.

            MiniZinc-specific implementation.

            Arguments:
                display:            either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                                    default/None: nothing displayed
                time_limit:         stop after this many seconds (default: None)
                solution_limit:     stop after this many solutions (default: None)
                call_from_model:    whether the method is called from a CPMpy Model instance or not
                **kwargs:           any keyword argument, sets parameters of solver object, overwrites construction-time kwargs

            Returns: 
                number of solutions found
        """
        # XXX: check that no objective function??
        if self.has_objective():
            raise NotSupportedError("Minizinc Python does not support finding all optimal solutions (yet)")

        import asyncio

        # set time limit
        if time_limit is not None:
            if time_limit <= 0:
                raise ValueError("Time limit must be positive")

        # HAD TO DEFINE OUR OWN ASYNC HANDLER
        coroutine = self._solveAll(display=display, time_limit=time_limit,
                                    solution_limit=solution_limit, **kwargs)
        # THE FOLLOWING IS STRAIGHT FROM `minizinc.instance.solve()`
        # LETS HOPE IT DOES NOT DIVERGE FROM UPSTREAM
        if sys.version_info >= (3, 7):
            if sys.platform == "win32":
                asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
            return asyncio.run(coroutine)
        else:
            if sys.platform == "win32":
                loop = asyncio.ProactorEventLoop()
            else:
                loop = asyncio.events.new_event_loop()

            try:
                asyncio.events.set_event_loop(loop)
                return loop.run_until_complete(coroutine)
            finally:
                asyncio.events.set_event_loop(None)
                loop.close()

    def minizinc_string(self) -> str:
        """
            Returns the model as represented in the Minizinc language.
        """
        return "".join(self._pre_solve()[1]._code_fragments)

    def flatzinc_string(self, **kwargs) -> str:
        """
            Returns the model as represented in the Flatzinc language.
        """
        with self._pre_solve()[1].flat(**kwargs) as (fzn, ozn, statistics):
            with open(fzn.name) as f:
                f.seek(0)
                contents = f.readlines()
        return "".join(contents)



"""
    CPMpy is a numpy-based library for conveniently modeling constraint programming problems in Python.

    Documentation in docs/index.rst
    as well as online at: https://cpmpy.readthedocs.io/

    Source code and bug reports at https://github.com/CPMpy/cpmpy

    The package constists of 4 modules:
    - `model`: a generic container for expressions (constraints and an objective), it can also search for an available solver and call it
    - `expressions`: all forms of expression objects that allow you to specify constraints and objectives over variables
    - `solvers`: CPMpy classes that translate a model into approriate calls of a solver's API
    - `transformations`: common methods for transforming expressions into other expressions, used by `solvers` modules to simplify/rewrite expressions
"""
# Tias Guns, 2019-2025

__version__ = "0.9.27"


from .expressions import *
from .model import Model
from .solvers.utils import SolverLookup



"""
    Interface to CP Optimizer's Python API.

    CP Optimizer, also a feature of IBM ILOG Optimization Studio, is a software library of constraint programming tools 
    supporting constraint propagation, domain reduction, and highly optimized solution search.

    Always use :func:`cp.SolverLookup.get("cpo") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'docplex' python package is installed:

    .. code-block:: console
    
        $ pip install docplex

    docplex documentation:
    https://ibmdecisionoptimization.github.io/docplex-doc/

    You will also need to install CPLEX Optimization Studio from IBM's website,
    and add the location of the CP Optimizer binary to your path.
    There is a free community version available.
    https://www.ibm.com/products/ilog-cplex-optimization-studio

    See detailed installation instructions at:
    https://www.ibm.com/docs/en/icos/22.1.2?topic=2212-installing-cplex-optimization-studio

    Academic license:
    https://community.ibm.com/community/user/ai-datascience/blogs/xavier-nodet1/2020/07/09/cplex-free-for-students

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
       :nosignatures:

        CPM_cpo
"""

import time
from typing import Optional
import warnings
import pkg_resources

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from .. import DirectConstraint
from ..expressions.core import Expression, Comparison, Operator, BoolVal
from ..expressions.globalconstraints import GlobalConstraint
from ..expressions.globalfunctions import GlobalFunction
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl, intvar
from ..expressions.utils import is_num, is_any_list, eval_comparison, argval, argvals, get_bounds
from ..transformations.get_variables import get_variables
from ..transformations.normalize import toplevel_list
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.safening import no_partial_functions



class CPM_cpo(SolverInterface):
    """
    Interface to CP Optimizer's Python API.

    Creates the following attributes (see parent constructor for more):

    - ``cpo_model``: object, CP Optimizers model object

    Documentation of the solver's own Python API: (all modeling functions)
    https://ibmdecisionoptimization.github.io/docplex-doc/cp/docplex.cp.modeler.py.html#module-docplex.cp.modeler

    """

    _docp = None  # Static attribute to hold the docplex.cp module

    @classmethod
    def get_docp(cls):
        if cls._docp is None:
            import docplex.cp as docp  # Import only once
            cls._docp = docp
        return cls._docp

    @staticmethod
    def supported():
        return CPM_cpo.installed() and CPM_cpo.license_ok()

    @staticmethod
    def installed():
        # try to import the package
        try:
            import docplex.cp as docp
            return True
        except ModuleNotFoundError:
            return False

    @staticmethod
    def license_ok():
        if not CPM_cpo.installed():
            warnings.warn(f"License check failed, python package 'docplex' is not installed! Please check 'CPM_cpo.installed()' before attempting to check license.")
            return False
        else:
            try:
                from docplex.cp.model import CpoModel
                mdl = CpoModel()
                mdl.solve(LogVerbosity='Quiet')
                return True
            except:
                return False
            
    @staticmethod
    def version() -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.

        For CPO, two version numbers get returned: ``<docplex version>/<solver version>``
        """
        try:
            import docplex.cp as docp
            s = docp.solver.solver.CpoSolver(docp.model.CpoModel())
            return f"{pkg_resources.get_distribution('docplex').version}/{s.get_solver_version()}"
        except (pkg_resources.DistributionNotFound, ModuleNotFoundError):
            return None

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
            cpm_model: Model(), a CPMpy Model() (optional)
            subsolver: str, name of a subsolver (optional)
        """
        if not self.installed():
            raise Exception("CPM_cpo: Install the python package 'docplex'")

        if not self.license_ok():
            raise Exception("You need to install the CPLEX Optimization Studio to use this solver. "
                            "Also make sure that the binary is in your path")

        docp = self.get_docp()
        assert subsolver is None
        self.cpo_model = docp.model.CpoModel()
        super().__init__(name="cpo", cpm_model=cpm_model)

    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.cpo_model
    
    def solve(self, time_limit=None, solution_callback=None, **kwargs):
        """
            Call the CP Optimizer solver

            Arguments:
                time_limit (float, optional):   maximum solve time in seconds 
                solution_callback (an `docplex.cp.solver.solver_listener.CpoSolverListener` object):   CPMpy includes its own, namely `CpoSolutionCounter`. If you want to count all solutions, 
                                                                                                        don't forget to also add the keyword argument 'enumerate_all_solutions=True'.
                kwargs:                         any keyword argument, sets parameters of solver object

            Arguments that correspond to solver parameters:


            =============================   ============
            Argument                        Description
            =============================   ============
            LogVerbosity                    Determines the verbosity of the search log. Choose a value from  ['Quiet', 'Terse', 'Normal', 'Verbose']. Default value is 'Quiet'.
            OptimalityTolerance             This parameter sets an absolute tolerance on the objective value for optimization models. The value is a positive float. Default value is 1e-09.
            RelativeOptimalityTolerance     This parameter sets a relative tolerance on the objective value for optimization models. The optimality of a solution is proven if either of the two parameters' criteria is fulfilled.
            Presolve                        This parameter controls the presolve of the model to produce more compact formulations and to achieve more domain reduction. Possible values for this parameter are On (presolve is activated) and Off (presolve is deactivated).
                                            The value is a symbol in ['On', 'Off']. Default value is 'On'.
            Workers                         This parameter sets the number of workers to run in parallel to solve your model. The value is a positive integer. Default value is Auto. (Auto = use all available CPU cores)
            =============================   ============

            All solver parameters are documented here: https://ibmdecisionoptimization.github.io/docplex-doc/cp/docplex.cp.parameters.py.html#docplex.cp.parameters.CpoParameters

        """
        docp = self.get_docp()

        # ensure all vars are known to solver        
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(intvar(1, 1) == 1)

        # call the solver, with parameters
        if 'LogVerbosity' not in kwargs:
            kwargs['LogVerbosity'] = 'Quiet'
        
        # set time limit
        if time_limit is not None and time_limit <= 0:
            raise ValueError("Time limit must be positive")
        
        # create solver object
        self.cpo_solver = docp.solver.solver.CpoSolver(
            self.cpo_model,
            TimeLimit=time_limit, 
            **kwargs, 
            listeners=[solution_callback] if solution_callback is not None else None
        )

        self.cpo_result = self.cpo_solver.solve()

        # new status, translate runtime
        self.cpo_status = self.cpo_result.get_solve_status()
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.cpo_result.get_solve_time() # wallclock time in (float) seconds

        # translate solver exit status to CPMpy exit status
        if self.cpo_status == "Feasible":
            self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif self.cpo_status == "Infeasible":
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif self.cpo_status == "Unknown":
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        elif self.cpo_status == "Optimal":
            self.cpm_status.exitstatus = ExitStatus.OPTIMAL
        elif self.cpo_status == "JobFailed":
            self.cpm_status.exitstatus = ExitStatus.ERROR
        elif self.cpo_status == "JobAborted":
            self.cpm_status.exitstatus = ExitStatus.NOT_RUN
        else:  # another?
            raise NotImplementedError(self.cpo_status)  # a new status type was introduced, please report on GitHub

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                if isinstance(cpm_var,_BoolVarImpl):
                    # because cp optimizer does not have boolean variables we use an integer var x with domain 0, 1
                    # and then replace a boolvar by x == 1
                    sol_var = sol_var.children[0]
                    cpm_var._value = bool(self.cpo_result.get_var_solution(sol_var).get_value())
                else:
                    cpm_var._value = self.cpo_result.get_var_solution(sol_var).get_value()

            # translate objective, for optimisation problems only
            if self.has_objective():
                self.objective_value_ = self.cpo_result.get_objective_value()

        else:  # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var.clear()

        return has_sol

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            A shorthand to (efficiently) compute all (optimal) solutions, map them to CPMpy and optionally display the solutions.

            If the problem is an optimization problem, returns only optimal solutions.

            Arguments:
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping.
                         Default is None, meaning nothing is displayed.
                time_limit: Stop after this many seconds. Default is None.
                solution_limit: Stop after this many solutions. Default is None.
                call_from_model: Whether the method is called from a CPMpy Model instance or not.
                **kwargs: Any other keyword arguments.

            Returns:
                int: Number of solutions found.
        """

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        # edge case, empty model, ensure the solver has something to solve
        if not len(self.user_vars):
            self.add(intvar(1, 1) == 1)

        docp = self.get_docp()
        solution_count = 0

        # TODO: convert to use 'start_search' and solution callback handlers
        start = time.time()
        while ((time_limit is None) or (time_limit > 0)) and self.solve(time_limit=time_limit, **kwargs):

            # display if needed
            if display is not None:
                if isinstance(display, Expression):
                    print(argval(display))
                elif isinstance(display, list):
                    print(argvals(display))
                else:
                    display()  # callback

            # count and stop
            solution_count += 1
            if solution_count == solution_limit:
                break

            if self.has_objective():
                # only find all optimal solutions
                self.cpo_model.add(self.cpo_model.get_objective_expression().children[0] == self.objective_value_)

            # add nogood on the user variables
            solvars = []
            vals = []
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                cpm_value = cpm_var._value
                if isinstance(cpm_var, _BoolVarImpl):
                    # because cp optimizer does not have boolean variables we use an integer var x with domain 0, 1
                    # and then replace a boolvar by x == 1
                    sol_var = sol_var.children[0]
                    cpm_value = int(cpm_value)
                solvars.append(sol_var)
                vals.append(cpm_value)
            self.cpo_model.add(docp.modeler.forbidden_assignments(solvars, [vals]))

            if time_limit is not None: # update remaining time
                time_limit -= self.status().runtime
        end = time.time()

        # update solver status
        self.cpm_status.runtime = end - start
        if solution_count:
            if solution_count == solution_limit:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
            elif self.cpm_status.exitstatus == ExitStatus.UNSATISFIABLE:
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        # else: <- is implicit since nothing needs to update
        #     if self.cpm_status.exitstatus == ExitStatus.UNSATISFIABLE:
        #         self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        #     elif self.cpm_status.exitstatus == ExitStatus.UNKNOWN:
        #         self.cpm_status.exitstatus = ExitStatus.UNKNOWN

        return solution_count

    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        docp = self.get_docp()
        if is_num(cpm_var): # shortcut, eases posting constraints
            return cpm_var

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            return (self.solver_var(cpm_var._bv) == 0)

        # create if it does not exist
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                # note that a binary var is an integer var with domain (0,1), you cannot do boolean operations on it.
                # we should add == 1 to turn it into a boolean expression
                revar = docp.expression.binary_var(str(cpm_var)) == 1
            elif isinstance(cpm_var, _IntVarImpl):
                revar = docp.expression.integer_var(min=cpm_var.lb, max=cpm_var.ub, name=str(cpm_var))
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            self._varmap[cpm_var] = revar
            self.cpo_model.add(revar >= cpm_var.lb) # ensure the model also has the variable

        # return from cache
        return self._varmap[cpm_var]

    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            ``objective()`` can be called multiple times, only the last one is stored

            .. note::
            
                technical side note: any constraints created during conversion of the objective are permanently posted to the solver
        """
        dom = self.get_docp().modeler
        if self.has_objective():
            self.cpo_model.remove(self.cpo_model.get_objective_expression())
        expr = self._cpo_expr(expr)
        if minimize:
            self.cpo_model.add(dom.minimize(expr))
        else:
            self.cpo_model.add(dom.maximize(expr))

    def has_objective(self):
        return self.cpo_model.get_objective() is not None

    # `add()` first calls `transform()`
    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the :ref:`Adding a new solver` docs on readthedocs for more information.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: list of Expression
        """
        # apply transformations
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = no_partial_functions(cpm_cons, safen_toplevel=frozenset({}))
        # count is only supported with a constant to be counted, so we decompose
        supported = {"alldifferent", 'inverse', 'nvalue', 'element', 'table', 'indomain',
                     "negative_table", "gcc", 'max', 'min', 'abs', 'cumulative', 'no_overlap'}
        supported_reified = {"alldifferent", 'table', 'indomain', "negative_table"} # global functions by default here
        cpm_cons = decompose_in_tree(cpm_cons, supported=supported, supported_reified=supported_reified, csemap=self._csemap)
        # no flattening required
        return cpm_cons

    def add(self, cpm_expr):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

            :param cpm_expr: CPMpy expression, or list thereof
            :type cpm_expr: Expression or list of Expression

            :return: self
        """

        # add new user vars to the set
        get_variables(cpm_expr, collect=self.user_vars)

        for cpm_con in self.transform(cpm_expr):
            # translate each expression tree, then post straight away
            cpo_con = self._cpo_expr(cpm_con)
            self.cpo_model.add(cpo_con)

        return self
    __add__ = add  # avoid redirect in superclass

    def _cpo_expr(self, cpm_con):
        """
            CP Optimizer supports nested expressions,
            so we recursively translate our expressions to theirs.

            Accepts single constraints or a list thereof, return type changes accordingly.

        """
        dom = self.get_docp().modeler
        if is_any_list(cpm_con):
            # arguments can be lists
            return [self._cpo_expr(con) for con in cpm_con]

        elif isinstance(cpm_con, BoolVal):
            return cpm_con.args[0]

        elif is_num(cpm_con):
            return cpm_con

        elif isinstance(cpm_con, _NumVarImpl):
            return self.solver_var(cpm_con)

        # Operators: base (bool), lhs=numexpr, lhs|rhs=boolexpr (reified ->)
        elif isinstance(cpm_con, Operator):
            arity, _ = Operator.allowed[cpm_con.name]
            # 'and'/n, 'or'/n, '->'/2
            if cpm_con.name == 'and':
                return dom.logical_and(self._cpo_expr(cpm_con.args))
            elif cpm_con.name == 'or':
                return dom.logical_or(self._cpo_expr(cpm_con.args))
            elif cpm_con.name == '->':
                return dom.if_then(*self._cpo_expr(cpm_con.args))
            elif cpm_con.name == 'not':
                return dom.logical_not(self._cpo_expr(cpm_con.args[0]))

            # 'sum'/n, 'wsum'/2
            elif cpm_con.name == 'sum':
                return dom.sum(self._cpo_expr(cpm_con.args))
            elif cpm_con.name == 'wsum':
                w = cpm_con.args[0]
                x = self._cpo_expr(cpm_con.args[1])
                return dom.scal_prod(w,x)

            # 'sub'/2, 'mul'/2, 'div'/2, 'pow'/2, 'm2od'/2
            elif arity == 2 or cpm_con.name == "mul":
                assert len(cpm_con.args) == 2, "Currently only support multiplication with 2 vars"
                x, y = self._cpo_expr(cpm_con.args)
                if cpm_con.name == 'sub':
                    return x - y
                elif cpm_con.name == "mul":
                    return x * y
                elif cpm_con.name == "div":
                    return x // y
                elif cpm_con.name == "pow":
                    return x ** y
                elif cpm_con.name == "mod":
                    return x % y
            # '-'/1
            elif cpm_con.name == "-":
                return -self._cpo_expr(cpm_con.args[0])

            else:
                raise NotImplementedError(f"Operator {cpm_con} not (yet) implemented for CP Optimizer, "
                                          f"please report on github if you need it")

        # Comparisons (just translate the subexpressions and re-post)
        elif isinstance(cpm_con, Comparison):
            lhs, rhs = self._cpo_expr(cpm_con.args)

            # post the comparison
            return eval_comparison(cpm_con.name, lhs, rhs)
        # rest: base (Boolean) global constraints
        elif isinstance(cpm_con, GlobalConstraint):
            if cpm_con.name == 'alldifferent':
                return dom.all_diff(self._cpo_expr(cpm_con.args))
            elif cpm_con.name == "gcc":
                vars, vals, occ = self._cpo_expr(cpm_con.args)
                cons = [dom.distribute(occ, vars, vals)]
                if cpm_con.closed:  # not supported by cpo, so post separately
                    cons += [dom.allowed_assignments(v, vals) for v in vars]
                return cons
            elif cpm_con.name == "inverse":
                x, y = self._cpo_expr(cpm_con.args)
                return dom.inverse(x, y)
            elif cpm_con.name == "table":
                arr, table = self._cpo_expr(cpm_con.args)
                return dom.allowed_assignments(arr, table)
            elif cpm_con.name == "indomain":
                expr, arr = self._cpo_expr(cpm_con.args)
                return dom.allowed_assignments(expr, arr)
            elif cpm_con.name == "negative_table":
                arr, table = self._cpo_expr(cpm_con.args)
                return dom.forbidden_assignments(arr, table)
            elif cpm_con.name == "cumulative":
                start, dur, end, height, capacity = cpm_con.args
                docp = self.get_docp()
                total_usage = []
                cons = []
                for s, d, e, h in zip(start, dur, end, height):
                    bounds_d = get_bounds(d)
                    # Special case for tasks with duration 0
                    # -> cpo immediately returns UNSAT if done through tasks
                    if bounds_d[1] == bounds_d[0] == 0:
                        cpo_s, cpo_e = self.solver_vars([s, e])
                        cons += [cpo_s == cpo_e] # enforce 0 duration
                        # no restrictions on height due to zero duration and thus no contribution to capacity
                        continue
                    # Normal setting
                    cpo_s, cpo_d, cpo_e, cpo_h = self.solver_vars([s, d, e, h])                   
                    task = docp.expression.interval_var(start=get_bounds(s), size=get_bounds(d), end=get_bounds(e))
                    task_height = dom.pulse(task, get_bounds(h))
                    cons += [dom.start_of(task) == cpo_s, dom.size_of(task) == cpo_d, dom.end_of(task) == cpo_e]
                    cons += [cpo_h == dom.height_at_start(task, task_height)]
                    total_usage.append(task_height)
                cons += [dom.sum(total_usage) <= self.solver_var(capacity)]
                return cons
            elif cpm_con.name == "no_overlap":
                start, dur, end  = cpm_con.args
                docp = self.get_docp()
                cons = []
                tasks = []
                for s, d, e in zip(start, dur, end):
                    cpo_s, cpo_d, cpo_e = self.solver_vars([s, d, e])
                    task = docp.expression.interval_var(start=get_bounds(s), size=get_bounds(d), end=get_bounds(e))
                    tasks.append(task)
                    cons += [dom.start_of(task) == cpo_s, dom.size_of(task) == cpo_d, dom.end_of(task) == cpo_e]
                return cons + [dom.no_overlap(tasks)]
            # a direct constraint, make with cpo (will be posted to it by calling function)
            elif isinstance(cpm_con, DirectConstraint):
                return cpm_con.callSolver(self, self.cpo_model)

            else:
                try:
                    cpo_global = getattr(dom, cpm_con.name)
                    return cpo_global(self._cpo_expr(cpm_con.args))  # works if our naming is the same
                except AttributeError:
                    raise ValueError(f"Global constraint {cpm_con} not known in CP Optimizer, please report on github.")

        elif isinstance(cpm_con, GlobalFunction):
            if cpm_con.name == "element":
                return dom.element(*self._cpo_expr(cpm_con.args))
            elif cpm_con.name == "min":
                return dom.min(self._cpo_expr(cpm_con.args))
            elif cpm_con.name == "max":
                return dom.max(self._cpo_expr(cpm_con.args))
            elif cpm_con.name == "abs":
                return dom.abs(self._cpo_expr(cpm_con.args)[0])
            elif cpm_con.name == "nvalue":
                return dom.count_different(self._cpo_expr(cpm_con.args))

        raise NotImplementedError("CP Optimizer: constraint not (yet) supported", cpm_con)


# solvers are optional, so this file should be interpretable
# even if cpo is not installed...
try:
    from docplex.cp.solver.solver_listener import CpoSolverListener
    import time

    class CpoSolutionCounter(CpoSolverListener):
        """
        Native CP Optimizer callback for solution counting.

        It is based on cpo's built-in `CpoSolverListener`.

        use with CPM_cpo as follows:

        .. code-block:: python
            
            cb = CpoSolutionCounter()
            s.solve(solution_callback=cb)

        then retrieve the solution count with ``cb.solution_count()``

        Arguments:
            verbose (bool, default: False): whether to print info on every solution found 
    """

        def __init__(self, verbose=False):
            super().__init__()
            self.__solution_count = 0
            self.__verbose = verbose
            if self.__verbose:
                self.__start_time = time.time()

        def result_found(self, solver, sres):
            """Called on each new solution."""
            if self.__verbose:
                current_time = time.time()
                obj = sres.get_objective_value()
                print('Solution %i, time = %0.2f s, objective = %i' %
                      (self.__solution_count, current_time - self.__start_time, obj))
            self.__solution_count += 1

        def solution_count(self):
            """Returns the number of solutions found."""
            return self.__solution_count

    class CpoSolutionPrinter(CpoSolutionCounter):
        """
            Native CP Optimizer callback for solution printing.

            Subclasses :class:`CpoSolutionCounter`, see those docs too.

            Use with :class:`CPM_cpo` as follows:

            .. code-block:: python

                cb = CpoSolutionPrinter(s, display=vars)
                s.solve(solution_callback=cb)

            For multiple variables (single or NDVarArray), use:
            ``cb = CpoSolutionPrinter(s, display=[v, x, z])``.

            For a custom print function, use for example:
            
            .. code-block:: python

                def myprint():
                    print(f"x0={x[0].value()}, x1={x[1].value()}")
                
                cb = CpoSolutionPrinter(s, printer=myprint)

            Optionally retrieve the solution count with ``cb.solution_count()``.

            Arguments:
                verbose (bool, default = False): whether to print info on every solution found 
                display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                            default/None: nothing displayed
                solution_limit (default = None): stop after this many solutions 
        """
        def __init__(self, solver, display=None, solution_limit=None, verbose=False):
            super().__init__(verbose)
            self._solution_limit = solution_limit
            # we only need the cpmpy->solver varmap from the solver
            self._varmap = solver._varmap
            # identify which variables to populate with their values
            self._cpm_vars = []
            self._display = display
            if isinstance(display, (list,Expression)):
                self._cpm_vars = get_variables(display)
            elif callable(display):
                # might use any, so populate all (user) variables with their values
                self._cpm_vars = solver.user_vars

        def result_found(self, solver, sres):
            """Called on each new solution."""
            if len(self._cpm_vars):
                # populate values before printing
                for cpm_var in self._cpm_vars:
                    # it might be an NDVarArray
                    if hasattr(cpm_var, "flat"):
                        for cpm_subvar in cpm_var.flat:
                            sol_var = self._varmap[cpm_subvar]
                            if isinstance(cpm_var,_BoolVarImpl):
                                sol_var = sol_var.children[0]
                                cpm_var._value = bool(sres.get_var_solution(sol_var).get_value())
                            else:
                                cpm_var._value = sres.get_var_solution(sol_var).get_value()
                    elif isinstance(cpm_var, _BoolVarImpl):
                        sol_var = self._varmap[cpm_subvar].children[0]
                        cpm_var._value = bool(sres.get_var_solution(sol_var).get_value())
                    else:
                        sol_var = self._varmap[cpm_subvar]
                        cpm_var._value = sres.get_var_solution(sol_var).get_value()

                if isinstance(self._display, Expression):
                    print(argval(self._display))
                elif isinstance(self._display, list):
                    # explicit list of expressions to display
                    print(argvals(self._display))
                else: # callable
                    self._display()

            # check for count limit
            if self.solution_count() == self._solution_limit:
                self.end_solve()

except ImportError:
    pass  # Ok, no cpo installed...



"""
Command-line interface for CPMpy.

This module provides a simple CLI to interact with CPMpy, primarily to display
version information about CPMpy itself and the available solver backends.

Usage:
    cpmpy <COMMAND>

Commands:
    version   Show the CPMpy library version and the versions of installed solver backends.
"""

import argparse
from cpmpy import __version__
import cpmpy as cp


def command_version(args):
    print(f"CPMpy version: {__version__}")
    cp.SolverLookup().print_version()

def main():
    parser = argparse.ArgumentParser(description="CPMpy command line interface")
    subparsers = parser.add_subparsers(dest="command", required=True)

    # cpmpy version
    version_parser = subparsers.add_parser("version", help="Show version information on CPMpy and its solver backends")
    version_parser.set_defaults(func=command_version)

    args = parser.parse_args()
    args.func(args)



'''
Custom exception classes, for finer grained error handling
'''


class CPMpyException(Exception):
    '''Parent class for all our exceptions'''
    pass


class MinizincPathException(CPMpyException):
    '''Raised when `minizinc` is not added to PATH'''
    pass

class MinizincNameException(CPMpyException):
    '''Raised when a variable is a keyword or otherwise violates Minizinc naming rules'''
    pass

class MinizincBoundsException(CPMpyException):
    '''Raised when an integer overflows MiniZinc's bounds of (-2147483646..2147483646)'''
    pass

class ChocoBoundsException(CPMpyException):
    '''Raised when an integer overflows Choco's integer bounds of (-2147483646..2147483646)'''
    pass

class NotSupportedError(CPMpyException):
    '''Raised when a solver does not support a certain feature'''
    pass

class IncompleteFunctionError(CPMpyException):
    '''Raised when an expression's value is not defined for its sub-expressions (e.g. `x div y` where `y` is assigned 0)'''
    pass

class TypeError(CPMpyException):
    '''Raised when an expression receives sub-expressions of the wrong type'''
    pass

class GCSVerificationException(CPMpyException):
    '''Raised when GCS fails proof logging by VeriPB'''
    pass

class TransformationNotImplementedError(CPMpyException):
    '''Raised when a transformation is not implemented for a certain expression'''
    pass


#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## utils.py
##
"""
Internal utilities for expression handling.

    =================
    List of functions
    =================
    .. autosummary::
        :nosignatures:

        is_bool
        is_int
        is_num
        is_false_cst
        is_true_cst
        is_boolexpr
        is_pure_list
        is_any_list
        is_transition
        flatlist
        all_pairs
        argval
        argvals
        eval_comparison
        get_bounds     
"""

import cpmpy as cp
import numpy as np
import math
from collections.abc import Iterable  # for flatten
from itertools import combinations
from cpmpy.exceptions import IncompleteFunctionError


def is_bool(arg):
    """ is it a boolean (incl numpy variants)
    """
    return isinstance(arg, (bool, np.bool_, cp.BoolVal))


def is_int(arg):
    """ can it be interpreted as an integer? (incl bool and numpy variants)
    """
    return isinstance(arg, (bool, np.bool_, cp.BoolVal, int, np.integer))


def is_num(arg):
    """ is it an int or float? (incl numpy variants)
    """
    return isinstance(arg, (bool, np.bool_, cp.BoolVal, int, np.integer, float, np.floating))


def is_false_cst(arg):
    """ is the argument the constant False (can be of type bool, np.bool and BoolVal)
    """
    if arg is False or arg is np.False_:
        return True
    elif isinstance(arg, cp.BoolVal):
        return not arg.value()
    return False


def is_true_cst(arg):
    """ is the argument the constant True (can be of type bool, np.bool and BoolVal)
    """
    if arg is True or arg is np.True_:
        return True
    elif isinstance(arg, cp.BoolVal):
        return arg.value()
    return False


def is_boolexpr(expr):
    """ is the argument a boolean expression or a boolean value
    """
    #boolexpr
    if hasattr(expr, 'is_bool'):
        return expr.is_bool()
    #boolean constant
    return is_bool(expr)


def is_pure_list(arg):
    """ is it a list or tuple?
    """
    return isinstance(arg, (list, tuple))


def is_any_list(arg):
    """ is it a list or tuple or numpy array?
    """
    return isinstance(arg, (list, tuple, np.ndarray))

def flatlist(args):
    """ recursively flatten arguments into one single list
    """
    return list(_flatten(args))


def _flatten(args):
    """ flattens the irregular nested list into an iterator

        from: https://stackoverflow.com/questions/2158395/flatten-an-irregular-list-of-lists
    """
    for el in args:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from _flatten(el)
        else:
            yield el


def all_pairs(args):
    """ returns all pairwise combinations of elements in args
    """
    return list(combinations(args, 2))


def argval(a):
    """ returns .value() of Expression, otherwise the variable itself

        We check with hasattr instead of isinstance to avoid circular dependency
    """
    if hasattr(a, "value"):
        try:
            val = a.value()
        except IncompleteFunctionError as e:
            if a.is_bool():
                return False
            else:
                raise e
    else:
        val = a

    if isinstance(val, np.generic):
        return val.item() # ensure it is a Python native value
    return val


def argvals(arr):
    if is_any_list(arr):
        return [argvals(arg) for arg in arr]
    return argval(arr)


def eval_comparison(str_op, lhs, rhs):
    """
        Internal function: evaluates the textual `str_op` comparison operator
        lhs <str_op> rhs

        Valid str_op's:
        * '=='
        * '!='
        * '>'
        * '>='
        * '<'
        * '<='

        Especially useful in decomposition and transformation functions that already involve a comparison.
    """
    if isinstance(lhs, (np.integer, np.bool_)):
        lhs = int(lhs)
    if isinstance(rhs, (np.integer, np.bool_)):
        rhs = int(rhs)

    if str_op == '==':
        return lhs == rhs
    elif str_op == '!=':
        return lhs != rhs
    elif str_op == '>':
        return lhs > rhs
    elif str_op == '>=':
        return lhs >= rhs
    elif str_op == '<':
        return lhs < rhs
    elif str_op == '<=':
        return lhs <= rhs
    else:
        raise Exception("Not a known comparison:", str_op)

def get_bounds(expr):
    """ return the bounds of the expression
    returns appropriately rounded integers
    """

    # import here to avoid circular import
    # from cpmpy.expressions.core import Expression
    # from cpmpy.expressions.variables import cpm_array

    if isinstance(expr, cp.expressions.core.Expression):
        return expr.get_bounds()
    elif is_any_list(expr):
        lbs, ubs = zip(*[get_bounds(e) for e in expr])
        return list(lbs), list(ubs) # return list as NDVarArray is covered above
    else:
        assert is_num(expr), f"All Expressions should have a get_bounds function, `{expr}`"
        if is_bool(expr):
            return int(expr), int(expr)
        return math.floor(expr), math.ceil(expr)

def implies(expr, other):
    """ like :func:`~cpmpy.expressions.core.Expression.implies`, but also safe to use for non-expressions """
    if isinstance(expr, cp.expressions.core.Expression):
        return expr.implies(other)
    elif is_true_cst(expr):
        return other
    elif is_false_cst(expr):
        return cp.BoolVal(True)
    else:
        return expr.implies(other)

# Specific stuff for ShortTabel global (should this be in globalconstraints.py instead?)
STAR = "*" # define constant here
def is_star(arg):
    """
        Check if arg is star as used in the ShortTable global constraint
    """
    return isinstance(arg, type(STAR)) and arg == STAR



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## model.py
##
"""
    The :class:`Model <cpmpy.model.Model>` class is a lazy container for constraints and an objective function.
    Constraints and objectives are CPMpy :mod:`expressions <cpmpy.expressions>`.

    It is lazy in that it only stores the constraints and objective that are added
    to it. Processing only starts when :meth:`solve() <cpmpy.model.Model.solve>` is called, and this does not modify
    the constraints or objective stored in the model.

    A model can be solved multiple times, and constraints can be added inbetween solve calls.
    Note that constraints are added using the ``+=`` operator (implemented by :meth:`__add__() <cpmpy.model.Model.__add__>`).

    See the full list of functions below.

    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:
        :toctree:

        Model
"""
import copy
import warnings

import numpy as np

from .exceptions import NotSupportedError
from .expressions.core import Expression
from .expressions.variables import NDVarArray
from .expressions.utils import is_any_list
from .solvers.utils import SolverLookup
from .solvers.solver_interface import SolverInterface, SolverStatus, ExitStatus

import pickle

class Model(object):
    """
    CPMpy Model object, contains the constraint and objective expressions
    """

    def __init__(self, *args, minimize=None, maximize=None):
        """
            Arguments of constructor:

            Arguments:
                `*args`: Expression object(s) or list(s) of Expression objects
                `minimize`: Expression object representing the objective to minimize
                `maximize`: Expression object representing the objective to maximize

            At most one of minimize/maximize can be set, if none are set, it is assumed to be a satisfaction problem
        """
        assert ((minimize is None) or (maximize is None)), "can not set both minimize and maximize"
        self.cpm_status = SolverStatus("Model") # status of solving this model, will be replaced

        # init list of constraints and objective
        self.constraints = []
        self.objective_ = None
        self.objective_is_min = None

        if len(args) == 1 and is_any_list(args):
            args = args[0]  # historical shortcut, treat as *args
        # use `__add__()` for typecheck
        if is_any_list(args):
            # add (and type-check) one by one
            for a in args:
                self += a
        else:
            self += args

        # store objective if present
        if maximize is not None:
            self.maximize(maximize)
        if minimize is not None:
            self.minimize(minimize)

        
    def add(self, con):
        """
        Add one or more constraints to the model.

        Arguments:
            con (Expression or list): Expression object(s) or list(s) of Expression objects representing constraints

        Returns:
            Model: Returns self to allow for method chaining

        Example:
            .. code-block:: python

                m = Model()
                m += [x > 0]
        """
        if is_any_list(con):
            # catch some beginner mistakes: check that top-level Expressions in the list have Boolean return type
            for elem in con:
                if isinstance(elem, Expression) and not elem.is_bool() and not isinstance(elem, NDVarArray):
                    raise Exception(f"Model error: constraints must be expressions that return a Boolean value, `{elem}` does not.")

            if len(con) == 0:
                # ignore empty list
                return self
            elif len(con) == 1:
                # unpack size 1 list
                con = con[0]

        elif isinstance(con, Expression) and not con.is_bool():
            # catch some beginner mistakes: ensure that a top-level Expression has Boolean return type
            raise Exception(f"Model error: constraints must be expressions that return a Boolean value, `{con}` does not.")

        self.constraints.append(con)
        return self
    __add__ = add  # Make __add__() (for the += operation) be the same as add() 


    def minimize(self, expr):
        """
            Minimize the given objective function

            `minimize()` can be called multiple times, only the last one is stored
        """
        self.objective(expr, minimize=True)

    def maximize(self, expr):
        """
            Maximize the given objective function

            `maximize()` can be called multiple times, only the last one is stored
        """
        self.objective(expr, minimize=False)

    def objective(self, expr, minimize):
        """
            Users will typically use :meth:`minimize() <cpmpy.model.Model.minimize>` or :meth:`maximize() <cpmpy.model.Model.maximize>` to set the objective function,
            this is the generic implementation for both.

            Arguments:
                expr (Expression):      the CPMpy expression that represents the objective function
                minimize (bool):        whether it is a minimization problem (True) or maximization problem (False)

            'objective()' can be called multiple times, only the last one is stored
        """
        self.objective_ = expr
        self.objective_is_min = minimize

    def has_objective(self):
        """
            Check if the model has an objective function

            Returns:
                bool: True if the model has an objective function, False otherwise
        """
        return self.objective_ is not None

    def objective_value(self):
        """
            Returns the value of the objective function of the last solver run on this model

            Returns:
                an integer or 'None' if it is not run or is a satisfaction problem
        """
        return self.objective_.value()

    def solve(self, solver=None, time_limit=None, **kwargs):
        """ Send the model to a solver and get the result.

            Run :func:`SolverLookup.solvernames() <cpmpy.solvers.SolverLookup.solvernames>` to find out the valid solver names on your system. (default: None = first available solver)

        Arguments:
            solver (string or a name in SolverLookup.solvernames() or a SolverInterface class (Class, not object!), optional): 
                name of a solver to use.
            time_limit (int or float, optional): time limit in seconds

            
        Returns:
            bool: the computed output:

            - True      if a solution is found (not necessarily optimal, e.g. could be after timeout)
            - False     if no solution is found
        """
        if kwargs and solver is None:
            raise NotSupportedError("Specify the solver when using kwargs, since they are solver-specific!")

        if isinstance(solver, SolverInterface):
            # for advanced use, call its constructor with this model
            s = solver(self)
        else:
            s = SolverLookup.get(solver, self)

        # call solver
        ret = s.solve(time_limit=time_limit, **kwargs)
        # store CPMpy status (s object has no further use)
        self.cpm_status = s.status()
        return ret

    def solveAll(self, solver=None, display=None, time_limit=None, solution_limit=None, **kwargs):
        """
            Compute all solutions and optionally display the solutions.

            If no solution is found, the solver status will be 'Unsatisfiable'.
            If at least one solution was found and the solver exhausted all possible solutions, the solver status will be 'Optimal', otherwise 'Feasible'.

            Arguments:
                display:            either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                                    default/None: nothing displayed
                solution_limit:     stop after this many solutions (default: None)

            Returns:
                int: number of solutions found (within the time and solution limit)
        """
        if kwargs and solver is None:
            raise NotSupportedError("Specify the solver when using kwargs, since they are solver-specific!")

        if isinstance(solver, SolverInterface):
            # for advanced use, call its constructor with this model
            s = solver(self)
        else:
            s = SolverLookup.get(solver, self)

        # call solver
        ret = s.solveAll(display=display,time_limit=time_limit,solution_limit=solution_limit, call_from_model=True, **kwargs)
        # store CPMpy status (s object has no further use)
        self.cpm_status = s.status()
        return ret

    def status(self):
        """
            Returns the status of the latest solver run on this model

            Status information includes exit status (optimality) and runtime.

            Returns:
                an object of :class:`SolverStatus`
        """
        return self.cpm_status

    def __repr__(self):
        """
            Returns a string representation of the model

            Returns:
                str: A string representation of the model
        """
        cons_str = ""
        for c in self.constraints:
            cons_str += "    {}\n".format(c)

        obj_str = ""
        if not self.objective_ is None:
            if self.objective_is_min:
                obj_str = "minimize "
            else:
                obj_str = "maximize "
        obj_str += str(self.objective_)
            
        return "Constraints:\n{}Objective: {}".format(cons_str, obj_str)


    def to_file(self, fname):
        """
            Serializes this model to a ``.pickle`` format

            Arguments:
                fname (FileDescriptorOrPath): Filename of the resulting serialized model
        """
        with open(fname,"wb") as f:
            pickle.dump(self, file=f)


    @staticmethod
    def from_file(fname):
        """
            Reads a Model instance from a binary pickled file

            Returns:
                an object of :class: `Model`
        """
        with open(fname, "rb") as f:
            m = pickle.load(f)
            # bug 158, we should increase the boolvar/intvar counters to avoid duplicate names
            from cpmpy.transformations.get_variables import get_variables_model  # avoid circular import
            from cpmpy.expressions.variables import _BoolVarImpl, _IntVarImpl, _BV_PREFIX, _IV_PREFIX # avoid circular import
            vs = get_variables_model(m)
            bv_counter = 0
            iv_counter = 0
            for v in vs:
                if v.name.startswith(_BV_PREFIX):
                    try:
                        bv_counter = max(bv_counter, int(v.name[2:])+1)
                    except:
                        pass
                elif v.name.startswith(_IV_PREFIX):
                    try:
                        iv_counter = max(iv_counter, int(v.name[2:])+1)
                    except:
                        pass

            if (_BoolVarImpl.counter > 0 and bv_counter > 0) or \
                    (_IntVarImpl.counter > 0 and iv_counter > 0):
                warnings.warn(f"from_file '{fname}': contains auxiliary {_IV_PREFIX}*/{_BV_PREFIX}* variables with the same name as already created. Only add expressions created AFTER loadig this model to avoid issues with duplicate variables.")
            _BoolVarImpl.counter = max(_BoolVarImpl.counter, bv_counter)
            _IntVarImpl.counter = max(_IntVarImpl.counter, iv_counter)
            return m

    def copy(self):
        """
            Makes a shallow copy of the model.
            Constraints and variables are shared among the original and copied model (references to the same Expression objects). The /list/ of constraints itself is different, so adding or removing constraints from one model does not affect the other.
        """
        if self.objective_is_min:
            return Model(self.constraints, minimize=self.objective_)
        else:
            return Model(self.constraints, maximize=self.objective_)



    # keep for backwards compatibility
    def deepcopy(self, memodict={}):
        warnings.warn("Deprecated, use copy.deepcopy() instead, will be removed in stable version", DeprecationWarning)
        return copy.deepcopy(self, memodict)



from setuptools import find_packages, setup
import codecs
import os.path

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

with open("README.md", "r", encoding="utf8") as readme_file:
    long_description = readme_file.read()


solver_dependencies = {
    "ortools": ["ortools"],
    "z3": ["z3-solver"],
    "choco": ["pychoco>=0.2.1"],
    "exact": ["exact>=2.1.0"],
    "minizinc": ["minizinc"],
    "pysat": ["python-sat"],
    "gurobi": ["gurobipy"],
    "pysdd": ["pysdd"],
    "gcs": ["gcspy"],
    "cpo": ["docplex"],
    "pumpkin": ["pumpkin-solver>=0.2.1"],
    "pindakaas": ["pindakaas>=0.2.1"],
    "cplex": ["docplex", "cplex"],
}
solver_dependencies["all"] = list({pkg for group in solver_dependencies.values() for pkg in group}) 

setup(
    name='cpmpy',
    version=get_version("cpmpy/__init__.py"),
    author='Tias Guns',
    author_email="tias.guns@kuleuven.be",
    license='Apache 2.0',
    description='A numpy-based library for modeling constraint programming problems',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CPMpy/cpmpy",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'ortools>=9.9',
        'numpy>=1.5',
        'setuptools',
    ],
    extras_require={
        # Solvers
        **solver_dependencies,
        # Tools
        "xcsp3": ["pycsp3", "requests", "tqdm", "matplotlib", "psutil", "filelock", "gnureadline; platform_system != 'Windows'", "pyreadline3; platform_system == 'Windows'"], # didn't add CLI-specific req since some are not cross-platform
        # Other
        "test": ["pytest", "pytest-timeout"],
        "docs": ["sphinx>=5.3.0", "sphinx_rtd_theme>=2.0.0", "myst_parser", "sphinx-automodapi", "readthedocs-sphinx-search>=0.3.2"],
    },
    entry_points={
        'console_scripts': [
            'cpmpy = cpmpy.cli:main',
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8'
)



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## TEMPLATE.py
##
"""
    Interface to TEMPLATE's API

    .. note::
        [GUIDELINE] Replace <TEMPLATE> by the solver's name, and implement the missing pieces
        The functions are ordered in a way that could be convenient to 
        start from the top and continue in that order.

    .. note::
        After you are done filling in the template, remove all comments starting with [GUIDELINE]

    .. warning::
        [GUIDELINE] do not include the python package at the top of the file,
        as CPMpy should also work without this solver installed.
        To ensure that, include it inside supported() and other functions that need it...

    <some information on the solver>

    Always use :func:`cp.SolverLookup.get("TEMPLATE") <cpmpy.solvers.utils.SolverLookup.get>` to instantiate the solver object.

    ============
    Installation
    ============

    Requires that the 'TEMPLATEpy' python package is installed:

    .. code-block:: console
    
        $ pip install TEMPLATEpy

    See detailed installation instructions at:
    <URL to detailed solver installation instructions, if any>

    The rest of this documentation is for advanced users.

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        CPM_template
"""

from typing import Optional
import warnings
import pkg_resources
from pkg_resources import VersionConflict

from .solver_interface import SolverInterface, SolverStatus, ExitStatus
from ..expressions.core import Expression, Comparison, Operator
from ..expressions.variables import _BoolVarImpl, NegBoolView, _IntVarImpl, _NumVarImpl
from ..expressions.utils import is_num, is_any_list, is_boolexpr
from ..transformations.get_variables import get_variables
from ..transformations.normalize import toplevel_list
from ..transformations.decompose_global import decompose_in_tree
from ..transformations.flatten_model import flatten_constraint
from ..transformations.comparison import only_numexpr_equality
from ..transformations.reification import reify_rewrite, only_bv_reifies

class CPM_template(SolverInterface):
    """
    Interface to TEMPLATE's API

    Creates the following attributes (see parent constructor for more):
    - tpl_model: object, TEMPLATE's model object

    Documentation of the solver's own Python API:
    <URL to docs or source code>
    """

    @staticmethod
    def supported():
        # try to import the package
        try:
            import TEMPLATEpy as gp
            # optionally enforce a specific version
            pkg_resources.require("TEMPLATEpy>=2.1.0")
            return True
        except ModuleNotFoundError: # if solver's Python package is not installed
            return False
        except VersionConflict: # unsupported version of TEMPLATEpy (optional)
            warnings.warn(f"CPMpy uses features only available from TEMPLATEpy version 0.2.1, "
                          f"but you have version {pkg_resources.get_distribution('TEMPLATEpy').version}.")
            return False
        except Exception as e:
            raise e

    @classmethod
    def version(cls) -> Optional[str]:
        """
        Returns the installed version of the solver's Python API.
        """
        try:
            return pkg_resources.get_distribution('TEMPLATEpy').version
        except pkg_resources.DistributionNotFound:
            return None
        
    # [GUIDELINE] If your solver supports different subsolvers, implement below method to return a list of subsolver names
    @staticmethod
    def solvernames(installed:bool=True):
        """
            Returns solvers supported by TEMPLATE (on your system).

            Arguments:
                installed (boolean): whether to filter the solvernames to those installed on your system (default True)
               
            Returns:
                list of solver names
        """
        if CPM_template.supported():
            # Collect solver names
            if installed:
                return # [ ... list of the installed subsolver names ... ]
            else:
                return # [ ... list of all subsolver names ... ]
        else:
            warnings.warn("TEMPLATE is not installed or not supported on this system.")
            return []

    # [GUIDELINE] If your solver supports different subsolvers, implement below method to return their respective versions
    @classmethod
    def solverversion(cls, subsolver:str) -> Optional[str]:
        """
        Returns the version of the requested subsolver.

        Arguments:
            subsolver (str): name of the subsolver

        Returns:
            Version number of the subsolver if installed, else None 
        """
        # return version of requested subsolver (if installed)
        # if requested subsolver does not exist, raise ValueError
        pass

    def __init__(self, cpm_model=None, subsolver=None):
        """
        Constructor of the native solver object

        Arguments:
        - cpm_model: Model(), a CPMpy Model() (optional)
        - subsolver: str, name of a subsolver (optional)
        """
        if not self.supported():
            raise Exception("CPM_TEMPLATE: Install the python package 'TEMPLATEpy' to use this solver interface.")

        import TEMPLATEpy

        assert subsolver is None # unless you support subsolvers, see pysat or minizinc

        # initialise the native solver object
        # [GUIDELINE] we commonly use 3-letter abbrivations to refer to native objects:
        #           OR-tools uses ort_solver, Gurobi grb_solver, Exact xct_solver...
        self.TPL_solver = TEMPLATEpy.Solver("cpmpy") 
        self.TPL_model = TEMPLATEpy.model("cpmpy")

        # initialise everything else and post the constraints/objective
        # [GUIDELINE] this superclass call should happen AFTER all solver-native objects are created.
        #           internally, the constructor relies on `add()` which uses the above solver native object(s)
        super().__init__(name="TEMPLATE", cpm_model=cpm_model)


    @property
    def native_model(self):
        """
            Returns the solver's underlying native model (for direct solver access).
        """
        return self.TPL_model

    def solve(self, time_limit=None, **kwargs):
        """
            Call the TEMPLATE solver

            Arguments:
            - time_limit:  maximum solve time in seconds (float, optional)
            - kwargs:      any keyword argument, sets parameters of solver object

            Arguments that correspond to solver parameters:
            # [GUIDELINE] Please document key solver arguments that the user might wish to change
            #       for example: assumptions=[x,y,z], log_output=True, var_ordering=3, num_cores=8, ...
            # [GUIDELINE] Add link to documentation of all solver parameters
        """

        # ensure all vars are known to solver
        self.solver_vars(list(self.user_vars))

        if time_limit is not None:
            self.TPL_solver.set_timelimit_seconds(time_limit)

        # [GUIDELINE] if your solver supports solving under assumptions, add `assumptions` as argument in header
        #       e.g., def solve(self, time_limit=None, assumptions=None, **kwargs):
        #       then translate assumptions here; assumptions are a list of Boolean variables or NegBoolViews

        # call the solver, with parameters
        my_status = self.TPL_solver.solve(**kwargs)
        # [GUIDELINE] consider saving the status as self.TPL_status so that advanced CPMpy users can access the status object.
        #       This is mainly useful when more elaborate information about the solve-call is saved into the status

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.TPL_solver.time() # wallclock time in (float) seconds

        # Translate solver exit status to CPMpy exit status
        # CSP:                         COP:
        # ├─ sat -> FEASIBLE           ├─ optimal -> OPTIMAL
        # ├─ unsat -> UNSATISFIABLE    ├─ sub-optimal -> FEASIBLE
        # └─ timeout -> UNKNOWN        ├─ unsat -> UNSATISFIABLE
        #                              └─ timeout -> UNKNOWN
        if my_status is True:
            # COP
            if self.has_objective():
                self.cpm_status.exitstatus = ExitStatus.OPTIMAL
            # CSP
            else:
                self.cpm_status.exitstatus = ExitStatus.FEASIBLE
        elif my_status is False:
            self.cpm_status.exitstatus = ExitStatus.UNSATISFIABLE
        elif my_status is None:
            # can happen when timeout is reached...
            self.cpm_status.exitstatus = ExitStatus.UNKNOWN
        else:  # another?
            raise NotImplementedError(my_status)  # a new status type was introduced, please report on github

        # True/False depending on self.cpm_status
        has_sol = self._solve_return(self.cpm_status)

        # translate solution values (of user specified variables only)
        self.objective_value_ = None
        if has_sol:
            # fill in variable values
            for cpm_var in self.user_vars:
                sol_var = self.solver_var(cpm_var)
                cpm_var._value = self.TPL_solver.value(sol_var)
                raise NotImplementedError("TEMPLATE: back-translating the solution values")

            # translate objective, for optimisation problems only
            if self.has_objective():
                self.objective_value_ = self.TPL_solver.ObjectiveValue()

        else: # clear values of variables
            for cpm_var in self.user_vars:
                cpm_var.clear()

        return has_sol


    def solver_var(self, cpm_var):
        """
            Creates solver variable for cpmpy variable
            or returns from cache if previously created
        """
        if is_num(cpm_var): # shortcut, eases posting constraints
            return cpm_var

        # [GUIDELINE] some solver interfaces explicitely create variables on a solver object
        #       then use self.TPL_solver.NewBoolVar(...) instead of TEMPLATEpy.NewBoolVar(...)

        # special case, negative-bool-view
        # work directly on var inside the view
        if isinstance(cpm_var, NegBoolView):
            return TEMPLATEpy.negate(self.solver_var(cpm_var._bv))

        # create if it does not exist
        if cpm_var not in self._varmap:
            if isinstance(cpm_var, _BoolVarImpl):
                revar = TEMPLATEpy.NewBoolVar(str(cpm_var))
            elif isinstance(cpm_var, _IntVarImpl):
                revar = TEMPLATEpy.NewIntVar(cpm_var.lb, cpm_var.ub, str(cpm_var))
            else:
                raise NotImplementedError("Not a known var {}".format(cpm_var))
            self._varmap[cpm_var] = revar

        # return from cache
        return self._varmap[cpm_var]


    # [GUIDELINE] if TEMPLATE does not support objective functions, you can delete this function definition
    def objective(self, expr, minimize=True):
        """
            Post the given expression to the solver as objective to minimize/maximize

            'objective()' can be called multiple times, only the last one is stored

            (technical side note: any constraints created during conversion of the objective

            are permanently posted to the solver)
        """
        # make objective function non-nested
        (flat_obj, flat_cons) = flatten_objective(expr, csemap=self._csemap)
        self += flat_cons # add potentially created constraints
        self.user_vars.update(get_variables(flat_obj)) # add objvars to vars

        # make objective function or variable and post
        obj = self._make_numexpr(flat_obj)
        # [GUIDELINE] if the solver interface does not provide a solver native "numeric expression" object,
        #         _make_numexpr may be removed and an objective can be posted as:
        #           self.TPL_solver.MinimizeWeightedSum(obj.args[0], self.solver_vars(obj.args[1]) or similar

        if minimize:
            self.TPL_solver.Minimize(obj)
        else:
            self.TPL_solver.Maximize(obj)

    def has_objective(self):
        return self.TPL_solver.hasObjective()

    def _make_numexpr(self, cpm_expr):
        """
            Converts a numeric CPMpy 'flat' expression into a solver-specific numeric expression

            Primarily used for setting objective functions, and optionally in constraint posting
        """

        # [GUIDELINE] not all solver interfaces have a native "numerical expression" object.
        #       in that case, this function may be removed and a case-by-case analysis of the numerical expression
        #           used in the constraint at hand is required in `add()`
        #       For an example of such solver interface, check out solvers/choco.py or solvers/exact.py

        if is_num(cpm_expr):
            return cpm_expr

        # decision variables, check in varmap
        if isinstance(cpm_expr, _NumVarImpl):  # _BoolVarImpl is subclass of _NumVarImpl
            return self.solver_var(cpm_expr)

        # any solver-native numerical expression
        if isinstance(cpm_expr, Operator):
           if cpm_expr.name == 'sum':
               return self.TPL_solver.sum(self.solver_vars(cpm_expr.args))
           elif cpm_expr.name == 'wsum':
               weights, vars = cpm_expr.args
               return self.TPL_solver.weighted_sum(weights, self.solver_vars(vars))
           # [GUIDELINE] or more fancy ones such as max
           #        be aware this is not the Maximum CONSTRAINT, but rather the Maximum NUMERICAL EXPRESSION
           elif cpm_expr.name == "max":
               return self.TPL_solver.maximum_of_vars(self.solver_vars(cpm_expr.args))
           # ...
        raise NotImplementedError("TEMPLATE: Not a known supported numexpr {}".format(cpm_expr))


    # `add()` first calls `transform()`
    def transform(self, cpm_expr):
        """
            Transform arbitrary CPMpy expressions to constraints the solver supports

            Implemented through chaining multiple solver-independent **transformation functions** from
            the `cpmpy/transformations/` directory.

            See the 'Adding a new solver' docs on readthedocs for more information.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: list of Expression
        """
        # apply transformations
        # XXX chose the transformations your solver needs, see cpmpy/transformations/
        cpm_cons = toplevel_list(cpm_expr)
        cpm_cons = decompose_in_tree(cpm_cons, supported={"alldifferent"})
        cpm_cons = flatten_constraint(cpm_cons)  # flat normal form
        cpm_cons = reify_rewrite(cpm_cons, supported=frozenset(['sum', 'wsum']))  # constraints that support reification
        cpm_cons = only_bv_reifies(cpm_cons)
        cpm_cons = only_numexpr_equality(cpm_cons, supported=frozenset(["sum", "wsum", "sub"]))  # supports >, <, !=
        # ...
        return cpm_cons

    def add(self, cpm_expr_orig):
        """
            Eagerly add a constraint to the underlying solver.

            Any CPMpy expression given is immediately transformed (through `transform()`)
            and then posted to the solver in this function.

            This can raise 'NotImplementedError' for any constraint not supported after transformation

            The variables used in expressions given to add are stored as 'user variables'. Those are the only ones
            the user knows and cares about (and will be populated with a value after solve). All other variables
            are auxiliary variables created by transformations.

        :param cpm_expr: CPMpy expression, or list thereof
        :type cpm_expr: Expression or list of Expression

        :return: self
        """

        # add new user vars to the set
        get_variables(cpm_expr_orig, collect=self.user_vars)

        # transform and post the constraints
        for cpm_expr in self.transform(cpm_expr_orig):

            if isinstance(cpm_expr, _BoolVarImpl):
                # base case, just var or ~var
                self.TPL_solver.add_clause([ self.solver_var(cpm_expr) ])

            elif isinstance(cpm_expr, Operator):
                if cpm_expr.name == "or":
                    self.TPL_solver.add_clause(self.solver_vars(cpm_expr.args))
                elif cpm_expr.name == "->": # half-reification
                    bv, subexpr = cpm_expr.args
                    # [GUIDELINE] example code for a half-reified sum/wsum comparison e.g. BV -> sum(IVs) >= 5
                    if isinstance(subexpr, Comparison):
                        lhs, rhs = subexpr.args
                        if isinstance(lhs, _NumVarImpl) or (isinstance(lhs, Operator) and lhs.name in {"sum", "wsum"}):
                            TPL_lhs = self._make_numexpr(lhs)
                            self.TPL_solver.add_half_reified_comparison(self.solver_var(bv),
                                                                        TPL_lhs, subexpr.name, self.solver_var(rhs))
                        else:
                            raise NotImplementedError("TEMPLATE: no support for half-reified comparison:", subexpr)
                    else:
                        raise NotImplementedError("TEMPLATE: no support for half-reified constraint:", subexpr)

            elif isinstance(cpm_expr, Comparison):
                lhs, rhs = cpm_expr.args

                # [GUIDELINE] == is used for both double reification and numerical comparisons
                #       need case by case analysis here. Note that if your solver does not support full-reification,
                #       you can rely on the transformation only_implies to convert all reifications to half-reification
                #       for more information, please reach out on github!
                if cpm_expr.name == "==" and is_boolexpr(lhs) and is_boolexpr(rhs): # reification
                    bv, subexpr = lhs, rhs
                    assert isinstance(lhs, _BoolVarImpl), "lhs of reification should be var because of only_bv_reifies"

                    if isinstance(subexpr, Comparison):
                        lhs, rhs = subexpr.args
                        if isinstance(lhs, _NumVarImpl) or (isinstance(lhs, Operator) and lhs.name in {"sum", "wsum"}):
                            TPL_lhs = self._make_numexpr(lhs)
                            self.TPL_solver.add_reified_comparison(self.solver_var(bv),
                                                                   TPL_lhs, subexpr.name, self.solver_var(rhs))
                        else:
                            raise NotImplementedError("TEMPLATE: no support for reified comparison:", subexpr)
                    else:
                        raise NotImplementedError("TEMPLATE: no support for reified constraint:", subexpr)

                # otherwise, numerical comparisons
                if isinstance(lhs, _NumVarImpl) or (isinstance(lhs, Operator) and lhs.name in {"sum", "wsum"}):
                    TPL_lhs = self._make_numexpr(lhs)
                    self.TPL_solver.add_comparison(TPL_lhs, cpm_expr.name, self.solver_var(rhs))
                # global functions
                elif cpm_expr.name == "==":
                    TPL_rhs = self.solver_var(rhs)
                    if lhs.name == "max":
                        self.TPL_solver.add_max_constraint(self.solver_vars(lhs), TPL_rhs)
                    elif lhs.name == "element":
                        TPL_arr, TPL_idx = self.solver_vars(lhs.args)
                        self.TPL_solver.add_element_constraint(TPL_arr, TPL_idx, TPL_rhs)
                    # elif...
                    else:
                        raise NotImplementedError("TEMPLATE: unknown equality constraint:", cpm_expr)
                else:
                    raise NotImplementedError("TEMPLATE: unknown comparison constraint", cpm_expr)

            # global constraints
            elif cpm_expr.name == "alldifferent":
                self.TPL_solver.add_alldifferent(self.solver_vars(cpm_expr.args))
            else:
                raise NotImplementedError("TEMPLATE: constraint not (yet) supported", cpm_expr)

        return self
    __add__ = add  # avoid redirect in superclass

    # Other functions from SolverInterface that you can overwrite:
    # solveAll, solution_hint, get_core

    def solveAll(self, display=None, time_limit=None, solution_limit=None, call_from_model=False, **kwargs):
        """
            A shorthand to (efficiently) compute all (optimal) solutions, map them to CPMpy and optionally display the solutions.

            If the problem is an optimization problem, returns only optimal solutions.

            Arguments:
                - display: either a list of CPMpy expressions, OR a callback function, called with the variables after value-mapping
                        default/None: nothing displayed
                - time_limit: stop after this many seconds (default: None)
                - solution_limit: stop after this many solutions (default: None)
                - call_from_model: whether the method is called from a CPMpy Model instance or not
                - any other keyword argument

            Returns: number of solutions found
        """

        # check if objective function (optional if solver doesn't support finding all solutions for COP)
        if self.has_objective():
            raise NotSupportedError("TEMPLATE does not support finding all optimal solutions")

        # A. Example code if solver supports callbacks
        if is_any_list(display):
            callback = lambda : print([var.value() for var in display])
        else:
            callback = display

        my_status = self.solve(time_limit, callback=callback, enumerate_all_solutions=True, **kwargs)

        # new status, translate runtime
        self.cpm_status = SolverStatus(self.name)
        self.cpm_status.runtime = self.TPL_solver.time() # wallclock time in (float) seconds (of entire solveAll call)

        # Translate solver exit status to CPMpy exit status
        # CSP & COP:
        # ├─ all solutions found -> OPTIMAL
        # ├─ at least 1 found (timeout / solution limit reached) -> FEASIBLE
        # ├─ 0 solutions found due to timeout -> UNKNOWN
        # └─ unsat -> UNSATISFIABLE

        # self.cpm_status.exitstatus = ...

        # clear user vars if no solution found
        if self.TPL_solver.SolutionCount() == 0:
            for var in self.user_vars:
                var.clear()
        return self.TPL_solver.SolutionCount()

        # B. Example code if solver does not support callbacks
        self.solve(time_limit, enumerate_all_solutions=True, **kwargs)
        solution_count = 0
        for solution in self.TPL_solver.GetAllSolutions():
            solution_count += 1
            # Translate solution to variables
            for cpm_var in self.user_vars:
                cpm_var._value = solution.value(solver_var)

            if display is not None:
                if isinstance(display, Expression):
                    print(display.value())
                elif isinstance(display, list):
                    print([v.value() for v in display])
                else:
                    display()  # callback

        # clear user vars if no solution found
        if solution_count == 0:
            for var in self.user_vars:
                var.clear()

        return solution_count



"""
  Meta-transformation for obtaining a CNF from a list of constraints.

  Converts the logical constraints into disjuctions using the tseitin transform,
  including flattening global constraints that are :func:`~cpmpy.expressions.core.Expression.is_bool()` and not in `supported`.

  .. note::
    The transformation is no longer used by the SAT solvers, and may be outdated.
    Check :meth:`CPM_pysat.transform <cpmpy.solvers.pysat.CPM_pysat.transform>` for an up-to-date alternative.
  
  Other constraints are copied verbatim so this transformation
  can also be used in non-pure CNF settings.

  The implementation first converts the list of constraints
  to **Flat Normal Form**, this already flattens subexpressions using
  auxiliary variables.

  What is then left to do is to tseitin encode the following into CNF:

  - ``BV`` with BV a ``BoolVar`` (or ``NegBoolView``)
  - ``or([BV])`` constraint
  - ``and([BV])`` constraint
  - ``BE != BV``  with ``BE :: BV|or()|and()|BV!=BV|BV==BV|BV->BV``
  - ``BE == BV``
  - ``BE -> BV``
  - ``BV -> BE``
"""
from ..expressions.core import Operator
from ..expressions.variables import _BoolVarImpl
from .reification import only_implies
from .flatten_model import flatten_constraint

def to_cnf(constraints, csemap=None):
    """
        Converts all logical constraints into **Conjunctive Normal Form**

        Arguments:
            constraints:    list[Expression] or Operator
            supported:      (frozen)set of global constraint names that do not need to be decomposed
    """
    fnf = flatten_constraint(constraints, csemap=csemap)
    fnf = only_implies(fnf, csemap=csemap)
    return flat2cnf(fnf)

def flat2cnf(constraints):
    """
        Converts from **Flat Normal Form** all logical constraints into **Conjunctive Normal Form**,
        including flattening global constraints that are :func:`~cpmpy.expressions.core.Expression.is_bool()` and not in `supported`.

        What is now left to do is to tseitin encode:

        - ``BV`` with BV a ``BoolVar`` (or ``NegBoolView``)
        - ``or([BV])`` constraint
        - ``and([BV])`` constraint
        - ``BE != BV``  with ``BE :: BV|or()|and()|BV!=BV|BV==BV|BV->BV``
        - ``BE == BV``
        - ``BE -> BV``
        - ``BV -> BE``

        We do it in a principled way for each of the cases. (in)equalities
        get transformed into implications, everything is modular.

        Arguments:
            constraints: list[Expression] or Operator
    """
    cnf = []
    for expr in constraints:
        # BE -> BE
        if expr.name == '->':
            a0,a1 = expr.args

            # BoolVar() -> BoolVar()
            if isinstance(a1, _BoolVarImpl) or \
                    (isinstance(a1, Operator) and a1.name == 'or'):
                cnf.append(~a0 | a1)
                continue

        # all other cases added as is...
        # TODO: we should raise here? is not really CNF...
        cnf.append(expr)

    return cnf



"""
    Normalize a toplevel list, or simplify Boolean expressions.
"""

import copy

import numpy as np
import cpmpy as cp

from ..expressions.core import BoolVal, Expression, Comparison, Operator
from ..expressions.globalfunctions import GlobalFunction
from ..expressions.utils import eval_comparison, is_false_cst, is_true_cst, is_boolexpr, is_num, is_bool
from ..expressions.variables import NDVarArray, _BoolVarImpl
from ..exceptions import NotSupportedError
from ..expressions.globalconstraints import GlobalConstraint


def toplevel_list(cpm_expr, merge_and=True):
    """
    Unravels nested lists and top-level AND's and ensures every element returned is a CPMpy Expression with :func:`~cpmpy.expressions.core.Expression.is_bool()` true.

    Arguments:
        cpm_expr:   Expression or list of Expressions
        merge_and:  if True then a toplevel 'and' will have its arguments merged at top level
    """

    # very efficient version with limited function lookups and list operations
    def unravel(lst, append):
        for e in lst:
            if isinstance(e, Expression):
                if isinstance(e, NDVarArray):  # sometimes does not have .name
                    unravel(e.flat, append)
                elif merge_and and e.name == "and":
                    unravel(e.args, append)
                else:
                    assert (e.is_bool()), f"Only boolean expressions allowed at toplevel, got {e}"
                    append(e) # presumably the most frequent case
            elif isinstance(e, (list, tuple, np.flatiter, np.ndarray)):
                unravel(e, append)
            elif e is False or e is np.False_:
                append(BoolVal(e))
            elif e is not True and e is not np.True_:  # if True: pass
                raise NotSupportedError(f"Expression {e} is not a valid CPMpy constraint")

    newlist = []
    append = newlist.append
    unravel((cpm_expr,), append)

    return newlist


def simplify_boolean(lst_of_expr, num_context=False):
    """
    Removes boolean constants from all CPMpy expressions, except for constants in global constraints/functions.
    Solver interfaces are expected to implement special cases of typing for global constraints themselves.

    Only resulting Boolean constant is literal 'false'.
    Boolean constants are promoted to `int` if in a numerical context,
    `ints` are never converted to `bool`.
    
    Arguments:
        list_of_expr: list of CPMpy expressions
    """

    newlist = []
    for expr in lst_of_expr:

        if isinstance(expr, Operator):
            if expr.has_subexpr():
                expr_args = simplify_boolean(expr.args, num_context=not expr.is_bool())
            else:
                expr_args = expr.args

            args = expr_args
            if expr.name == "or":
                i = 0
                while i < len(args):
                    a = args[i]
                    if isinstance(a, _BoolVarImpl):
                        i += 1 # default path
                    elif is_true_cst(a):
                        break
                    elif is_false_cst(a):
                        if args is expr_args: # will remove this one, need to copy args...
                            args = args.copy()
                        args.pop(i)
                    else:
                        i += 1

                if i != len(args): # found "True" along the way, early exit
                    newlist.append(1 if num_context else BoolVal(True))
                elif args is not expr.args: # removed something, or changed due to subexpr
                    newexpr = copy.copy(expr)
                    newexpr.update_args(args)
                    newlist.append(newexpr)
                else: # no changes
                    newlist.append(expr)

            elif expr.name == "and":
                # filter out True/False constants
                i = 0
                while i < len(args):
                    a = args[i]
                    if isinstance(a, _BoolVarImpl):
                        i += 1 # default path
                    elif is_false_cst(a):
                        break
                    elif is_true_cst(a):
                        if args is expr_args:  # will remove this one, need to copy args...
                            args = args.copy()
                        args.pop(i)
                    else:  # subexpression, should not happen here...
                        i += 1

                if i != len(args): # found "False" along the way, early exit
                    newlist.append(0 if num_context else BoolVal(False))
                elif args is not expr.args: # removed something, or changed due to subexpr
                    newexpr = copy.copy(expr)
                    newexpr.update_args(args)
                    newlist.append(newexpr)
                else: # no changes
                    newlist.append(expr)

            elif expr.name == "->":
                cond, bool_expr = args
                if is_false_cst(cond) or is_true_cst(bool_expr):
                    newlist.append(1 if num_context else BoolVal(True))
                elif is_true_cst(cond):
                    newlist.append(bool_expr)
                elif is_false_cst(bool_expr):
                    newlist += simplify_boolean([cp.transformations.negation.recurse_negation(cond)])
                else:
                    newlist.append(cond.implies(bool_expr))

            elif expr.name == "not":
                if isinstance(args[0], _BoolVarImpl): # fast path
                    newlist.append(~args[0])
                elif is_true_cst(args[0]):
                    newlist.append(0 if num_context else BoolVal(False))
                elif is_false_cst(args[0]):
                    newlist.append(1 if num_context else BoolVal(True))
                elif args is not expr.args:
                    expr = copy.copy(expr)
                    expr.update_args(expr_args)
                    newlist.append(expr)
                else: # nothing changed
                    newlist.append(expr)

            # numerical expressions
            elif expr.name == "wsum":
                weights, vars = args
                newvars = [int(v) if is_bool(v) else v for v in vars]
                if args is not expr.args or any(v1 is not v2 for v1,v2 in zip(vars, newvars)):
                    newexpr = copy.copy(expr)
                    newexpr.update_args([weights, newvars])
                    newlist.append(newexpr)
                else:
                    newlist.append(expr)

            else: # default case, check if anything is Bool that should be int
                newargs = [int(a) if is_bool(a) else a for a in args]
                if expr_args is not expr.args or any(v1 is not v2 for v1, v2 in zip(args, newargs)):
                    newexpr = copy.copy(expr)
                    newexpr.update_args(newargs)
                    newlist.append(newexpr)
                else:
                    newlist.append(expr)

        elif isinstance(expr, Comparison):
            lhs, rhs = simplify_boolean(expr.args, num_context=True)
            name = expr.name
            if is_num(lhs) and is_boolexpr(rhs):  # flip arguments of comparison to reduct nb of cases
                if name == "<":    name = ">"
                elif name == ">":  name = "<"
                elif name == "<=": name = ">="
                elif name == ">=": name = "<="
                lhs, rhs = rhs, lhs
            """
            Simplify expressions according to this table:
            x  | <0	 0  0.n	 1	>1
            ----------------------
            == |  F	~x	 F   x	 F
            != |  T	 x	 T  ~x	 T
            >  |  T	 x	 x   F	 F
            <  |  F	 F	~x  ~x	 T
            >= |  T	 T	 x   x	 F   
            <= |  F	~x	~x   T	 T
            """
            if is_boolexpr(lhs) and is_num(rhs):
                # direct simplification of boolean comparisons
                if isinstance(rhs, BoolVal):
                    rhs = int(rhs.value())
                if rhs < 0:
                    newlist.append(BoolVal(name in  {"!=", ">", ">="})) # all other operators evaluate to False
                elif rhs == 0:
                    if name == "!=" or name == ">":
                        newlist.append(lhs)
                    if name == "==" or name == "<=":
                        newlist.append(cp.transformations.negation.recurse_negation(lhs))
                    if name == "<":
                        newlist.append(0 if num_context else BoolVal(False))
                    if name == ">=":
                        newlist.append(1 if num_context else BoolVal(True))
                elif 0 < rhs < 1:
                    # a floating point value
                    if name == "==":
                        newlist.append(0 if num_context else BoolVal(False))
                    if name == "!=":
                        newlist.append(1 if num_context else BoolVal(True))
                    if name == "<" or name == "<=":
                        newlist.append(cp.transformations.negation.recurse_negation(lhs))
                    if name == ">" or name == ">=":
                        newlist.append(lhs)
                elif rhs == 1:
                    if name == "==" or name == ">=":
                        newlist.append(lhs)
                    if name == "!=" or name == "<":
                        newlist.append(cp.transformations.negation.recurse_negation(lhs))
                    if name == ">":
                        newlist.append(0 if num_context else BoolVal(False))
                    if name == "<=":
                        newlist.append(1 if num_context else BoolVal(True))
                elif rhs > 1:
                    newlist.append(BoolVal(name in {"!=", "<", "<="})) # all other operators evaluate to False
            else:
                res = eval_comparison(name, lhs, rhs)
                if is_bool(res): # Result is a Boolean constant
                    newlist.append(int(res) if num_context else BoolVal(res))
                else: # Result is an expression
                    newlist.append(res)                    
        elif isinstance(expr, (GlobalConstraint, GlobalFunction)):
            newargs = simplify_boolean(expr.args) # TODO: how to determine which are Bool/int?
            if any(a1 is not a2 for a1,a2 in zip(expr.args, newargs)):
                expr = copy.copy(expr)
                expr.update_args(newargs)
            newlist.append(expr)
        elif is_bool(expr):  # unlikely base-case (Boolean constant)
            newlist.append(int(expr) if num_context else BoolVal(expr))
        elif isinstance(expr, list):  # nested list in args (like for wsum)
            newlist.append(simplify_boolean(expr))
        else:  # variables/constants/direct constraints
            newlist.append(expr)
    return newlist



"""
Flattening a model (or individual constraints) into 'flat normal form'.

In flat normal form, constraints belong to one of three families with all arguments
either constants, variables, list of constants or list of variables, and
some binary constraints have a canonical order of variables.

Furthermore, it is 'negated normal' meaning that the ~ (negation operator) only appears
before a Boolean variable (in CPMpy, absorbed in a 'NegBoolView'),
and it is 'negation normal' meaning that the - (negative operator) only appears before
a constant, that is a - b :: a + -1*b :: wsum([1,-1],[a,b])

The three families of possible constraints are:

Base constraints: (no nesting)
------------------------------

=============================  ====================================  ==============
Boolean variable               ``Var``                                                                                                                                                    
Boolean operators              ``and([Var])``, ``or([Var])``         :class:`~cpmpy.expressions.core.Operator`, :func:`~cpmpy.expressions.core.Operator.is_bool()`                        
Boolean implication            ``Var -> Var``                        :class:`~cpmpy.expressions.core.Operator`, :func:`~cpmpy.expressions.core.Operator.is_bool()`                                                                                
Boolean equality               ``Var == Var``, ``Var == Constant``   :class:`~cpmpy.expressions.core.Comparison`                                                                          
Global constraint (Boolean)    ``global([Var]*)``                    :class:`~cpmpy.expressions.globalconstraints.GlobalConstraint`, :func:`~cpmpy.expressions.core.Operator.is_bool()`                                           
=============================  ====================================  ==============

Comparison constraints: (up to one nesting on one side)
-------------------------------------------------------

===============================  ============================================  ==============
Numeric equality                 ``Numexpr == Var``, ``Numexpr == Constant``   :class:`~cpmpy.expressions.core.Comparison`
Numeric disequality              ``Numexpr != Var``, ``Numexpr != Constant``   :class:`~cpmpy.expressions.core.Comparison`
Numeric inequality (>=,>,<,<=)   ``Numexpr >=< Var``                           :class:`~cpmpy.expressions.core.Comparison`
===============================  ============================================  ==============                                                    

**Numexpr:**

==================================================  ======================================  ==============
Operator (non-Boolean) with all args Var/constant   ``+``, ``*``, ``/``, ``mod``, ``wsum``  :class:`~cpmpy.expressions.core.Operator`, not :func:`~cpmpy.expressions.core.Operator.is_bool()`                      
Global constraint (non-Boolean)                     ``Max``, ``Min``, ``Element``           :class:`~cpmpy.expressions.globalconstraints.GlobalConstraint`, not :func:`~cpmpy.expressions.core.Operator.is_bool()` 
==================================================  ======================================  ==============

**wsum:**

.. todo::
    wsum([Const],[Var]) represents sum([Const]*[Var]) # TODO: not implemented yet

Reify/imply constraint: (up to two nestings on one side)
--------------------------------------------------------
=================================  =========================================  ==============
Reification (double implication)   ``Boolexpr == Var``                        :class:`~cpmpy.expressions.core.Comparison`                                                   
Implication                        ``Boolexpr -> Var``, ``Var -> Boolexpr``   :class:`~cpmpy.expressions.core.Operator`, :func:`~cpmpy.expressions.core.Operator.is_bool()` 
=================================  =========================================  ==============

**Boolexpr:**

==================================  =============================  ==============
Boolean operators                   ``and([Var])``, ``or([Var])``  :class:`~cpmpy.expressions.core.Operator`, not :func:`~cpmpy.expressions.core.Operator.is_bool()`                      
Boolean equality                    ``Var == Var``                 :class:`~cpmpy.expressions.core.Comparison`                                                                            
Global constraint (Boolean)         ``global([Var]*)``             :class:`~cpmpy.expressions.globalconstraints.GlobalConstraint`, :func:`~cpmpy.expressions.core.Operator.is_bool()`     
Comparison constraint (see above)                                  :class:`~cpmpy.expressions.core.Comparison`       
==================================  =============================  ==============                                                                     

    
Reification of a comparison is the most complex case as it can allow up to 3 levels of nesting in total, e.g.:

- (wsum([1,2,3],[IV1,IV2,IV3]) > 5) == BV
- (IV1 == IV2) == BV
- (BV1 == BV2) == BV3

Objective: (up to one nesting)
------------------------------
======================  ========================================  ============
Type                    Example                                   Notes                                                  
======================  ========================================  ============
Satisfaction problem    ``None``                                                                                         
Decision variable       ``Var``                                   :class:`~cpmpy.expressions.core.Operator`, name `sum`  
Linear                  ``sum([Var])``, ``wsum([Const],[Var])``   :class:`~cpmpy.expressions.core.Operator`, name `wsum` 
======================  ========================================  ============


The output after calling :func:`flatten_model()` or :func:`flatten_constraint()` will ONLY contain expressions
of the form specified above.

The flattening does not promise to do common subexpression elimination or to automatically group
commutative expressions (``and``, ``or``, ``sum``, ``wsum``, ...) but such optimisations should be added later.

.. todo::
    TODO: update behind_the_scenes.rst doc with the new 'flat normal form'
    TODO: small optimisations, e.g. and/or chaining (potentially after negation), see test_flatten
"""
import math
import builtins
import cpmpy as cp

from .normalize import toplevel_list, simplify_boolean
from ..expressions.core import *
from ..expressions.core import _wsum_should, _wsum_make
from ..expressions.variables import _NumVarImpl, _IntVarImpl, _BoolVarImpl
from ..expressions.utils import is_num, is_any_list, is_int, is_star
from .negation import push_down_negation


def flatten_model(orig_model, csemap=None):
    """
        Receives model, returns new model where every constraint is in 'flat normal form'
    """

    # the top-level constraints
    basecons = flatten_constraint(orig_model.constraints, csemap=csemap)

    # the objective
    if orig_model.objective_ is None:
        return cp.Model(*basecons)  # no objective, satisfaction problem
    else:
        (newobj, newcons) = flatten_objective(orig_model.objective_, csemap=csemap)
        basecons += newcons
        if orig_model.objective_is_min:
            return cp.Model(*basecons, minimize=newobj)
        else:
            return cp.Model(*basecons, maximize=newobj)


def flatten_constraint(expr, csemap=None):
    """
        input is any expression; except is_num(), pure _NumVarImpl,
        or Operator/GlobalConstraint with not is_bool()
        
        output: see definition of 'flat normal form' above.

        it will return 'Exception' if something is not supported

        .. todo::
            TODO, what built-in python error is best?
            RE TODO: we now have custom NotImpl/NotSupported
    """

    newlist = []
    # for backwards compatibility reasons, we now consider it a meta-
    # transformation, that calls (preceding) transformations itself
    # e.g. `toplevel_list()` ensures it is a list
    lst_of_expr = toplevel_list(expr)               # ensure it is a list
    lst_of_expr = push_down_negation(lst_of_expr)   # push negation into the arguments to simplify expressions
    lst_of_expr = simplify_boolean(lst_of_expr)     # simplify boolean expressions, and ensure types are correct
    for expr in lst_of_expr:

        if not expr.has_subexpr():
            newlist.append(expr)  # no need to do anything
            continue

        elif isinstance(expr, Operator):
            """
            - Base Boolean operators: and([Var]), or([Var])        (CPMpy class 'Operator', is_bool())
            - Base Boolean impliciation: Var -> Var                (CPMpy class 'Operator', is_bool())
            - Implication: Boolexpr -> Var                         (CPMpy class 'Operator', is_bool())
                           Var -> Boolexpr                         (CPMpy class 'Operator', is_bool())
            """
            # does not type-check that arguments are bool... Could do now with expr.is_bool()!
            if expr.name == 'or':
                # rewrites that avoid auxiliary var creation, should go to normalize?
                # in case of an implication in a disjunction, merge in
                if builtins.any(isinstance(a, Operator) and a.name == '->' for a in expr.args):
                    newargs = list(expr.args)  # take copy
                    for i,a in enumerate(newargs):
                        if isinstance(a, Operator) and a.name == '->':
                            newargs[i:i+1] = [~a.args[0],a.args[1]]
                    # there could be nested implications
                    newlist.extend(flatten_constraint(Operator('or', newargs), csemap=csemap))
                    continue
                # conjunctions in disjunctions could be split out by applying distributivity,
                # but this would explode the number of constraints in favour of having less auxiliary variables.
                # Testing has proven that this is not worth it.
            elif expr.name == '->':
                # some rewrite rules that avoid creating auxiliary variables
                # 1) if rhs is 'and', split into individual implications a0->and([a11..a1n]) :: a0->a11,...,a0->a1n
                if expr.args[1].name == 'and':
                    a1s = expr.args[1].args
                    a0 = expr.args[0]
                    newlist.extend(flatten_constraint([a0.implies(a1) for a1 in a1s], csemap=csemap))
                    continue
                # 2) if lhs is 'or' then or([a01..a0n])->a1 :: ~a1->and([~a01..~a0n] and split
                elif expr.args[0].name == 'or':
                    a0s = expr.args[0].args
                    a1 = expr.args[1]
                    newlist.extend(flatten_constraint([(~a1).implies(~a0) for a0 in a0s], csemap=csemap))
                    continue
                # 2b) if lhs is ->, like 'or': a01->a02->a1 :: (~a01|a02)->a1 :: ~a1->a01,~a1->~a02
                elif expr.args[0].name == '->':
                    a01,a02 = expr.args[0].args
                    a1 = expr.args[1]
                    newlist.extend(flatten_constraint([(~a1).implies(a01), (~a1).implies(~a02)], csemap=csemap))
                    continue

                # ->, allows a boolexpr on one side
                elif isinstance(expr.args[0], _BoolVarImpl):
                    # LHS is var, ensure RHS is normalized 'Boolexpr'
                    lhs,lcons = expr.args[0], ()
                    rhs,rcons = normalized_boolexpr(expr.args[1], csemap=csemap)
                else:
                    # make LHS normalized 'Boolexpr', RHS must be a var
                    lhs,lcons = normalized_boolexpr(expr.args[0], csemap=csemap)
                    rhs,rcons = get_or_make_var(expr.args[1], csemap=csemap)

                newlist.append(Operator(expr.name, (lhs,rhs)))
                newlist.extend(lcons)
                newlist.extend(rcons)
                continue



            # if none of the above cases + continue matched:
            # a normalizable boolexpr
            (con, flatcons) = normalized_boolexpr(expr, csemap=csemap)
            newlist.append(con)
            newlist.extend(flatcons)

        elif isinstance(expr, Comparison):
            """
    - Base Boolean equality: Var == Var                         (CPMpy class 'Comparison')
                             Var == Constant                    (CPMpy class 'Comparison')
    - Numeric equality:  Numexpr == Var                    (CPMpy class 'Comparison')
                         Numexpr == Constant               (CPMpy class 'Comparison')
    - Numeric disequality: Numexpr != Var                  (CPMpy class 'Comparison')
                           Numexpr != Constant             (CPMpy class 'Comparison')
    - Numeric inequality (>=,>,<,<=,): Numexpr >=< Var     (CPMpy class 'Comparison')
    - Reification (double implication): Boolexpr == Var    (CPMpy class 'Comparison')
            """
            exprname = expr.name  # so it can be modified
            lexpr, rexpr = expr.args
            rewritten = False

            # rewrite 'Var # Expr' to normalized 'Expr # Var' (where # is any comparator)
            if __is_flat_var(lexpr) and not __is_flat_var(rexpr):
                assert (expr.name in ('==', '!=', '>', '>=', '<', '<='))
                lexpr, rexpr = rexpr, lexpr
                rewritten = True

                # flip comparator in case of inequality
                if exprname == '>':
                    exprname = '<'
                elif exprname == '>=':
                    exprname = '<='
                elif exprname == '<':
                    exprname = '>'
                elif exprname == '<=':
                    exprname = '>='

            # already flat?
            if not expr.has_subexpr():
                if not rewritten:
                    newlist.append(expr)  # original
                else:
                    newlist.append(Comparison(exprname, lexpr, rexpr))
                continue

            # ensure rhs is var
            (rvar, rcons) = get_or_make_var(rexpr, csemap=csemap)
            # Reification (double implication): Boolexpr == Var
            # normalize the lhs (does not have to be a var, hence we call normalize instead of get_or_make_var
            if exprname == '==' and lexpr.is_bool():
                if rvar.is_bool():
                    # this is a reification
                    (lhs, lcons) = normalized_boolexpr(lexpr, csemap=csemap)
                else:
                    # integer comparison
                    (lhs, lcons) = get_or_make_var(lexpr, csemap=csemap)
            else:
                (lhs, lcons) = normalized_numexpr(lexpr, csemap=csemap)

            newlist.append(Comparison(exprname, lhs, rvar))
            newlist.extend(lcons)
            newlist.extend(rcons)

        elif isinstance(expr, cp.expressions.globalconstraints.GlobalConstraint):
            """
    - Global constraint: global([Var]*)          (CPMpy class 'GlobalConstraint')
            """
            (con, flatcons) = normalized_boolexpr(expr, csemap=csemap)
            newlist.append(con)
            newlist.extend(flatcons)

        else:
            # any other case (e.g. DirectConstraint), pass as is
            newlist.append(expr)

    return newlist


def flatten_objective(expr, supported=frozenset(["sum", "wsum"]), csemap=None):
    """
    - Decision variable: Var
    - Linear: 
        ======================                       ========
        sum([Var])                                   (CPMpy class 'Operator', name 'sum')
        wsum([Const],[Var])                          (CPMpy class 'Operator', name 'wsum')
        ======================                       ========
    """
    # lets be very explicit here
    if is_any_list(expr):
        # one source of errors is sum(v) where v is a matrix, use v.sum() instead
        raise Exception(f"Objective expects a single variable/expression, not a list of expressions: {expr}")

    expr = simplify_boolean([expr])[0]
    (flatexpr, flatcons) = normalized_numexpr(expr, csemap=csemap)  # might rewrite expr into a (w)sum
    if isinstance(flatexpr, Expression) and flatexpr.name in supported:
        return (flatexpr, flatcons)
    else:
        # any other numeric expression,
        var, cons = get_or_make_var(flatexpr, csemap=csemap)
        return (var, cons+flatcons)


def __is_flat_var(arg):
    """ True if the variable is a numeric constant, or a _NumVarImpl (incl subclasses)
    """
    return is_num(arg) or isinstance(arg, _NumVarImpl)

def __is_flat_var_or_list(arg):
    """ True if the variable is a numeric constant, or a _NumVarImpl (incl subclasses)
        or a list of __is_flat_var_or_list, or it is a wildcard as used in the ShortTable global constraint
    """
    return is_num(arg) or isinstance(arg, _NumVarImpl) or \
           is_any_list(arg) and all(__is_flat_var_or_list(el) for el in arg) or \
           is_star(arg)

def get_or_make_var(expr, csemap=None):
    """
        Must return a variable, and list of flat normal constraints
        Determines whether this is a Boolean or Integer variable and returns
        the equivalent of: (var, normalize(expr) == var)
    """

    if __is_flat_var(expr):
        return (expr, [])

    if is_any_list(expr):
        raise Exception(f"Expected single variable, not a list for: {expr}")

    if csemap is not None and expr in csemap:
        return csemap[expr], []

    if expr.is_bool():
        # normalize expr into a boolexpr LHS, reify LHS == bvar
        (flatexpr, flatcons) = normalized_boolexpr(expr, csemap=csemap)

        if isinstance(flatexpr,_BoolVarImpl):
            # avoids unnecessary bv == bv or bv == ~bv assignments
            return flatexpr,flatcons
        bvar = _BoolVarImpl()

        # save expr in dict
        if csemap is not None:
            csemap[expr] = bvar
        return bvar, [flatexpr == bvar] + flatcons

    else:
        # normalize expr into a numexpr LHS,
        # then compute bounds and return (newintvar, LHS == newintvar)
        (flatexpr, flatcons) = normalized_numexpr(expr, csemap=csemap)

        lb, ub = flatexpr.get_bounds()
        if not is_int(lb) or not is_int(ub):
            warnings.warn(f"CPMpy only uses integer variables, but found expression ({expr}) with domain {lb}({type(lb)}"
                          f" - {ub}({type(ub)}. CPMpy will rewrite this constriants with integer bounds instead.")
            lb, ub = math.floor(lb), math.ceil(ub)
        ivar = _IntVarImpl(lb, ub)

        # save expr in dict
        if csemap is not None:
            csemap[expr] = ivar
        return ivar, [flatexpr == ivar] + flatcons

def get_or_make_var_or_list(expr, csemap=None):
    """ Like get_or_make_var() but also accepts and recursively transforms lists
        Used to convert arguments of globals
    """

    if __is_flat_var_or_list(expr):
        return (expr,[])
    elif is_any_list(expr):
        flatvars, flatcons = zip(*[get_or_make_var(arg, csemap=csemap) for arg in expr])
        return (flatvars, [c for con in flatcons for c in con])
    else:
        return get_or_make_var(expr, csemap=csemap)


def normalized_boolexpr(expr, csemap=None):
    """
        input is any Boolean (is_bool()) expression
        output are all 'flat normal form' Boolean expressions that can be 'reified', meaning that

        .. code-block:: text

            subexpr == BoolVar
            subexpr -> BoolVar

        are valid output expressions.

        Currently, this is the case for subexpr:

        =====================================  =============================  ==============
        Boolean operators                      ``and([Var])``, ``or([Var])``  :class:`~cpmpy.expressions.core.Operator`, not :func:`~cpmpy.expressions.core.Operator.is_bool()`                      
        Boolean equality                       ``Var == Var``                 :class:`~cpmpy.expressions.core.Comparison`                                                                            
        Global constraint (Boolean)            ``global([Var]*)``             :class:`~cpmpy.expressions.globalconstraints.GlobalConstraint`, :func:`~cpmpy.expressions.core.Operator.is_bool()`     
        Comparison constraint (see elsewhere)                                 :class:`~cpmpy.expressions.core.Comparison`       
        =====================================  =============================  ==============                                                                     

        Result:
            (base_expr, base_cons) with:
            
            - base_expr: same as 'expr', but all arguments are variables
            - base_cons: list of flat normal constraints
    """
    assert(not __is_flat_var(expr))
    assert(expr.is_bool()) 

    if isinstance(expr, Operator):
        # and, or, ->

        # apply De Morgan's transform for "implies"
        if expr.name == '->':
            # TODO, optimisation if args0 is an 'and'?
            (lhs,lcons) = get_or_make_var(expr.args[0], csemap=csemap)
            # TODO, optimisation if args1 is an 'or'?
            (rhs,rcons) = get_or_make_var(expr.args[1], csemap=csemap)
            return ((~lhs | rhs), lcons+rcons)
        if expr.name == 'not':
            flatvar, flatcons = get_or_make_var(expr.args[0], csemap=csemap)
            return (~flatvar, flatcons)
        if not expr.has_subexpr():
            return (expr, [])
        else:
            # one of the arguments is not flat, flatten all
            flatvars, flatcons = zip(*[get_or_make_var(arg, csemap=csemap) for arg in expr.args])
            newexpr = Operator(expr.name, flatvars)
            return (newexpr, [c for con in flatcons for c in con])

    elif isinstance(expr, Comparison):
        if (expr.name != '!=') and (not expr.has_subexpr()):
            return (expr, [])  # shortcut
        else:
            # LHS can be boolexpr, RHS has to be variable

            lexpr, rexpr = expr.args
            exprname = expr.name

            # ==,!=: can swap if lhs is var and rhs is not
            if (exprname == '==' or exprname == '!=') and \
                not __is_flat_var(rexpr) and __is_flat_var(lexpr):
                lexpr, rexpr = rexpr, lexpr

            # ensure rhs is var
            (rvar, rcons) = get_or_make_var(rexpr, csemap=csemap)

            # LHS: check if Boolexpr == smth:
            if (exprname == '==' or exprname == '!=') and lexpr.is_bool():
                # this is a reified constraint, so lhs must be var too to be in normal form
                (lhs, lcons) = get_or_make_var(lexpr, csemap=csemap)
                if expr.name == '!=' and rvar.is_bool():
                    # != not needed, negate RHS variable
                    rvar = ~rvar
                    exprname = '=='
            else:
                # other cases: LHS is numexpr
                (lhs, lcons) = normalized_numexpr(lexpr, csemap=csemap)

            return (Comparison(exprname, lhs, rvar), lcons+rcons)

    else:
        """
        - Global constraint (Boolean): global([Var]*)          (CPMpy class 'GlobalConstraint', is_bool())
        """
        # just recursively flatten args, which can be lists
        if not expr.has_subexpr():
            return (expr, [])
        else:
            # recursively flatten all children
            flatargs, flatcons = zip(*[get_or_make_var_or_list(arg, csemap=csemap) for arg in expr.args])

            # take copy, replace args
            newexpr = copy.copy(expr) # shallow or deep? currently shallow
            newexpr.update_args(flatargs)
            return (newexpr, [c for con in flatcons for c in con])


def normalized_numexpr(expr, csemap=None):
    """
        all 'flat normal form' numeric expressions...

        Currently, this is the case for:

        - Operator (non-Boolean) with all args Var/constant (examples: +,*,/,mod,wsum)
                                                           (CPMpy class 'Operator', not is_bool())
        - Global constraint (non-Boolean) (examples: Max,Min,Element)
                                                           (CPMpy class 'GlobalConstraint', not is_bool()))

        Result:
            (base_expr, base_cons) with:
            
            - base_expr: same as 'expr', but all arguments are variables
            - base_cons: list of flat normal constraints
    """
    # XXX a boolexpr is also a valid numexpr... e.g. 30*(iv > 5) + ... see mario obj.
    if __is_flat_var(expr):
        return (expr, [])

    elif expr.is_bool():
        # unusual case, but its truth-value is a valid numexpr
        # so reify and return the boolvar
        return get_or_make_var(expr, csemap=csemap)

    elif isinstance(expr, Operator):
        # rewrite -a, const*a and a*const into a weighted sum, so it can be used as objective
        if expr.name == '-' or (expr.name == 'mul' and _wsum_should(expr)):
            return normalized_numexpr(Operator("wsum", _wsum_make(expr)), csemap=csemap)

        if not expr.has_subexpr():
            return (expr, [])

        # pre-process sum, to fold in nested subtractions and const*Exprs, e.g. x - y + 2*(z+r)
        if expr.name == "sum" and \
           all(isinstance(a, Expression) for a in expr.args) and \
           any((a.name == "-" or _wsum_should(a)) for a in expr.args):
            we = [_wsum_make(a) for a in expr.args]
            w = [wi for w,_ in we for wi in w]
            e = [ei for _,e in we for ei in e]
            return normalized_numexpr(Operator("wsum", (w,e)), csemap=csemap)

        # wsum needs special handling because expr.args is a tuple of which only 2nd one has exprs
        if expr.name == 'wsum':
            weights, sub_exprs  = expr.args
            # while here, avoid creation of auxiliary variables for compatible operators -/sum/wsum
            i = 0
            while(i < len(sub_exprs)): # can dynamically change
                if isinstance(sub_exprs[i], Operator) and \
                    ((sub_exprs[i].name in ['-', 'sum'] and
                      all(isinstance(a, Expression) for a in sub_exprs[i].args)) or
                     (sub_exprs[i].name == 'wsum' and
                      all(isinstance(a, Expression) for a in sub_exprs[i].args[1]))):  # TODO: avoid constants for now...
                    w,e = _wsum_make(sub_exprs[i])
                    # insert in place, and next iteration over same 'i' again
                    weights[i:i+1] = [weights[i]*wj for wj in w]
                    sub_exprs[i:i+1] = e
                else:
                    i = i+1

            # now flatten the resulting subexprs
            flatvars, flatcons = map(list, zip(*[get_or_make_var(arg, csemap=csemap) for arg in sub_exprs])) # also bool, reified...
            newexpr = Operator(expr.name, (weights, flatvars))
            return (newexpr, [c for con in flatcons for c in con])

        else: # generic operator
            # recursively flatten all children
            flatvars, flatcons = zip(*[get_or_make_var(arg, csemap=csemap) for arg in expr.args])

            newexpr = Operator(expr.name, flatvars)
            return (newexpr, [c for con in flatcons for c in con])
    else:
        # Globalfunction (examples: Max,Min,Element)

        # just recursively flatten args, which can be lists
        if not expr.has_subexpr():
            return (expr, [])
        else:
            # recursively flatten all children
            flatvars, flatcons = zip(*[get_or_make_var_or_list(arg, csemap=csemap) for arg in expr.args])

            # take copy, replace args
            newexpr = copy.copy(expr) # shallow or deep? currently shallow
            newexpr.update_args(flatvars)
            return (newexpr, [c for con in flatcons for c in con])
